# ls.py
# Functions to support DVR+ LinearSchedule calls
# Copyright June 2020, Timothy Goihl

from time import time, gmtime, strftime, mktime
import json
import copy
import logging
from datetime import timedelta, datetime
from dateutil import parser
import sys
import time


# Calls we need to make to SD
from sd import SDGetStationSchedule
from sd import SDGetProgramDetails
from sv import GetServiceList
from ut import LocalTimeStr

# The is the message header DVR+ expects when reuesting a schedule
# The Linear Schedule Result Header, a python dictionary
lsr = {
  "LinearScheduleResult": {
    # String GUID
    "RequestId": "6be49fd4-46b7-4429-acc2-1d385a441864",
    "TimeStamp": "2020-05-02T16:04:06.396Z",
    "Status": "PNE",  #PNE Processed No Errors
    "Errors": [],
    "EndTimestamp": "0001-01-01T00:00:00",
    "Build": "Development-v9-dev-9.15.08-173-r34606-34606",
    "Schedule": {
      "Locale": "en-US",
      "ServiceId": 12345,
      "Name": "Madison Area Channels - Broadcast TV",
      "ServiceType": "Broadcast",
      "StartDate": "2020-05-02T16:00:00Z",
      "Duration": 240,
      "TimeZones": [
        {
          "StartDateTime": "2020-03-08T08:00:00Z",
          "EndDateTime": "2020-11-01T07:00:00Z",
          "Offset": -300
        }
      ],
      "Airings": []    # empty list of airingDefs
    }
  }
}        

# airingDef
# Definition of the one 'program airing' object.  Fields will be updated with SD data
# Each display slot on the DVR+ Guide is one of these.  I doubt that DVr+ uses it all.
# DVR+ probably copies out the few things it needs and discards the rest.

COPY_MAX_CHARS = 512    # measured as 512 chars by expiriment

airingDef = {
      "ProgramId": "21085335",  # <10 chars?, seems to work with SD 12-char Program ID's, luckily. 
      "SeriesId":"1234",        # not used or changed
      "ServiceId": 12345,       # must match service id 99999 everywhere
      "SourceId": 11833,        # must match SourceID in services (StationID)
      "ParentNetworkId": 2,     # this will be set to the channels 'order' for a unique #
      "Channel": "3.1",
      "CallLetters": "WISCHD",
      "SourceDisplayName": "xx",
      "SourceLongName": "xx",
      "Title": "Prm Title", 
      "Copy": "~512 chars.",    # measured as 512 chars by experiment
      "AiringTime": "",         # 2020-05-02T05:00:00Z
      "Duration": 1440,         #DVR+ wants mins, SD supplies seconds, /60
      "AiringType": "Unknown",  # this is the magic 'New' status that allows 'record all new programs'
      "TVRating": "None",       # TV-14 etc
      "IconAvailable": False,
      "EpisodeTitle": "",
      "CC": False,              # These are the defaults, if SD keys are found
      "Stereo": False,          # they will be updated
      "SAP": False,
      "Dolby": False,
      "DVS": False,
      "HD": False,
      "HDLevel": "HD Level Unknown",  # HD Level Unknown,480p,480p,720p,1080i,1080p
        # I dont mess with any of these below this line
      "InstanceId": None,
      "UnknownInstance": False,
      "UPId": "autofill",
      "Color": "Color",
      "LetterBox": False,
      "ShowingType": "Unknown",  #Finale,Paid For,Premiere,Season Finale,Season Premiere,Series Finale,Series Premiere,Unknown
      "DSS": False,
      "Sports": False,
      "Category": "Other",
      "Subcategory": "other",
      "TiVoSupported": True,
      "DvbTriplet": "",
      "SL": "false",            # green text "false" in wireshark from CM
      "Subtitles": "false"      # green text "false" in wireshark from CM
    }


from ut import total_size
#
# Build a complete schedule by downloading the full SD schedule
# for all of the users favorite stations.  The downloaded SD data  
# will be reformatted into the json format required by the DVR+ and then
# added to a airing time sorted list.  The sorted list will be easy to 
# traverse for GetLinearSchedule to return data to DVR+.

# This will be a time-ordered list of SD programs reformatted into
# DVR+ json Airings format.  Actually its stored internally as a Python dictionary
# but converting a dict to json is 1 line of code.

from operator import itemgetter  #, attrgetter  # for sorting channels/airings into time order

# The airings or programs list will be built from SD data and then kept in memory.
# A request for a scheudle does not go out to SD in realtime, it just uses this cached list
# The MasterAiringList will be updated on demand when it is 'Stale'

MasterAiringList = []
LastBuiltTime = 0
BuildProgress = 0


#*****************************************************************************
# Key Mapping between various json objects
#
# ServicesResult    ServiceDetailResult   Linear Schedule Result     SD
# 
#  ServiceID
#
# Reference: SD API
# https://github.com/SchedulesDirect/JSON-Service/wiki/API-20141201
#
# BuildLinearSchedule creates the MasterAiringList[]

BuildInfo = {
  'Schedule Build Info': {
    "DaysLimit": 0,
    "BuildStartTime": "",
    "Build_End_Time": "",
    "Build_Progress": "",
    "Programs" : 0,
    "SchedMBs" : "",
    "SchdEnds" : "",
  }
}

# for pigstatus
def GetLSBuildInfo() :

    # Create the Section key if it does not exist
    if 'Schedule Build Info' not in BuildInfo :
        BuildInfo['Schedule Build Info'] = {}

    return BuildInfo


def BuildLinearSchedule(PiGScfg, serviceID, AbortEvent):

    global MasterAiringList
    global LastBuiltTime
    global DictOfDVRs
    global BuildInfo

    # get list of users Favorite Stations / channels 
    # it is a dict, the same list sent to DVR+ when it requests it in Message Type 2
    stationList = GetServiceList()

    if not stationList :
        logging.warning("LS:BLS: There are no configured Stations! Nothing to do!")
        return

    # for computing % complete.  +1 for computing the final size
    TotalChannels = stationList[-1]["Order"] + 1   # = len(stationList)

    BuildStartTime = datetime.utcnow()
    logging.info("LS:BLS: Start Time UTC: %s", BuildStartTime.strftime("%Y-%m-%dT%H:%M:%SZ"))
    
    # Status Grouping Title key
    #BuildInfo['Schedule Build Info'] = {}
    #BuildInfo['Schedule Build Info']['SchedulIsStale'] = ScheduleIsStale()

    # check if there is a DaysLimit in the config file (limit data days for testing)
    DaysLimit = PiGScfg.get('DaysLimit', 0)
    if DaysLimit :
        logging.info("LS:BLS: Limiting SD request to %d days(plus yesterday).", DaysLimit)
        
    BuildInfo['Schedule Build Info']['DaysLimit'] = DaysLimit

    ts,tz = LocalTimeStr( BuildStartTime,True )
    BuildInfo['Schedule Build Info']['BuildStartTime'] = ts
    BuildInfo['Schedule Build Info']["Build_End_Time"] = "Rebuilding..."   # Building now
    BuildInfo['Schedule Build Info']['Build_Progress'] = "0%"
    BuildInfo['Schedule Build Info']["Programs"] = 0
    BuildInfo['Schedule Build Info']["SchedMBs"] = ""
    BuildInfo['Schedule Build Info']["SchdEnds"] = ""

    # build a fresh new list - complete refresh   
    MasterAiringList.clear()
    
    # build a list of program ID as we go. will make getting descriptions faster
    PgmIdList = []
    DupIdList = []  # for debug
    
    # program airings will be built in a temp list, then be added to the master Airing list
    # we do this so we can add the program details just to this partially complete list
    # of airings
    tempList = []
    
    iconnames = ['None(0)','Schedule Direct Icons(1)','CM/Rovi Icons(2)','???(3)','???(4)'] 
    #Description Enhancements?
    logging.info("LS:BLS: CAST List : %s", PiGScfg['AppendCast'])
    logging.info("LS:BLS: Icon Type : %s", iconnames[PiGScfg['IconType']])
    logging.info("LS:BLS: Episd Info: %s", PiGScfg['PrependEpisode'])
    
    #==================================================================================
    # Loop through the station/channel objects in the Services List
    # The stationList has the Channel ID integers we need to get data for.
    
    BuildAbort = False   # Method to break out of Build process
    
    for aStation in stationList :

        if BuildAbort :  # Method to break out of Build process if a processnig error occurs below
            break   

        # Check for Abort Event, and convert it into a Build Abort
        if AbortEvent.is_set() :
            BuildAbort = True
            #AbortEvent.clear()  Builder will clear it
            break
        
        #time.sleep(10)  #xxx debug testing, slow down the build to see how dvrs respond

        # SD treats Station ID as a string, CM wants an integer
        
        iSourceId = aStation["SourceId"]  #int  DVR+ call stations SourceId as an integer
        
        # skip channels marked as 0 so that the channels can exist in the Service Lineup
        # This allows new stations to have an image icon before the schedule is available
        
        if iSourceId == 0 : continue
        
        # Part1 ========================================================================
        # Get the full Schedule for 1 station from Schedules Direct
        # Then scan thru SD's json and build a DVR+ compatible 'airings' list
        
        # SDGetStationSchedule tries to get yestday plus all days forward (limited by DaysLimit)
        # the returned data is all dates available for 1 station
        # DVR+ can only use 14 days of data, so ndays should be called DaysLimit as 14
        #print('st sch ', iSourceId)
        code, schj = SDGetStationSchedule( str(iSourceId), DaysLimit )
        #print('st sch ', iSourceId)
        
        # A 200 return, provides either an error block
        #[
        #  {
        #    "stationID": "25124",
        #    "serverID": "20141201.web.1",
        #    "code": 7020,
        #    "response": "SCHEDULE_RANGE_EXCEEDED",
        #    "minDate": null,
        #    "maxDate": null,
        #    "requestedDate": "2020-09-22",
        #    "message": "Date requested (2020-09-22) not within -> for stationID 25124."
        #  }
        #]
        # or program data
        #[
        #  {
        #    "stationID": "20454",
        #    "programs": [
        #            { prog keys },
        #    ],
        #  }
        #]
        #
        

        if code != 200 :
            logging.warning("LS:BLS: SDGetStationSchedule() Error(%d)- Not built!")
            MasterAiringList.clear()
            BuildAbort = True   # Method to break out of Build process
            break
        
        # clear the temp lists we use on each interation
        PgmIdList.clear()
        DupIdList.clear()
        tempList.clear()
 
        # SD Schedule data is in multiple blocks by date with a stationID header
        # There is also a differnt type of error header with a non-zero code.
        # we dont know which we will get

        for aSdStationId in schj :  # aSdStationId for Station Date element, find the stationID's
            
            #check for an error code, good blocks dont have an error code
            # get() does not fail. if 'code' not available, its sched data, default to 0=ok
            code = aSdStationId.get('code',0)
            
            # check for error code in json data
            # its probably a 7020 no-date error since I tried to get yesterday
            # yesterday exists for a while (and is needed around 7pm cst)
            if code > 0 :
                logging.debug("LS:Build Found code %d", code)
                continue;
             
            # Get Refernces to the 3 sections of the schedule data
            # ToDo, do i need try/catch or use Get() to prevent key error crashes?
            
            sids= aSdStationId['stationID']  # SD stationID or the Airing SourceId
            prl = aSdStationId['programs']   # programs list
            met = aSdStationId['metadata']   # metadata, md5 stuff
            
            # for 1 date in the returned list, parse the programs
            
            for aProgram in prl :  # aProgram scan the SD Programs results List
                
                # append a copy of a new airing object (the object format DVR+ needs)
                tempList.append( copy.deepcopy(airingDef) )

                #look for the showing 'new' key - does not always exist
                # Attempt to load SD new key, default to Not new
                newflag= aProgram.get('new', False)
                # Convert to DVR+ 'New'
                # DVR+ text must be one of these strings: Unknown New Live Repeat Taped
                newstr = "New" if newflag else "Unknown"  # ternary assignment
                tempList[-1]["AiringType"]= newstr
               
                #ToDo what if any these keys do not exist
                
                #Copy what we can from the SD Schedule items
                #need to make a list of programID's for Stage 2
                #because SD schedule has no program details - must be requested separately
                
                pidstr = copy.copy( aProgram["programID"] )  #programID is a string
                                
                # build list of program ID strings for this channel for Part 2
                if pidstr not in PgmIdList :    # only add PID's that are not duplicates
                    PgmIdList.append(pidstr)
                elif pidstr not in DupIdList :
                    DupIdList.append(pidstr)    # curious, what are the dups typically?

                tempList[-1]["ProgramId"] = pidstr
                    
                tempList[-1]["ParentNetworkId"] = aStation['ParentNetworkId'];               
                
                #DVR+ requires the channel number in each Airing
                tempList[-1]["Channel"]   = aStation['Channel']   # like 3.1, copy from config
                tempList[-1]["ServiceId"] = serviceID    # 99999 must match in airings
                tempList[-1]["SourceId"]  = iSourceId    # SourceID must match service header
                # Airing time format matches directly - one thing I dont have to modify!!
                tempList[-1]["AiringTime"]= copy.copy( aProgram["airDateTime"])
                # DVR+ expects minutes, SD supplies seconds
                tempList[-1]["Duration"]  = copy.copy( int(aProgram["duration"]/60) )
                
                # copy from Settings file info about a station
                tempList[-1]["CallLetters"]       = aStation["CallLetters"]
                tempList[-1]["SourceDisplayName"] = aStation["DisplayName"]
                tempList[-1]["SourceLongName"]    = aStation["FullName"]

                # If SD has audio properties.... they are optional
                aplst = aProgram.get("audioProperties")  
                if aplst :
                    #logging.info('audio %s', aplst)
                    tempList[-1]["CC"]     = 'cc' in aplst   # True shows CC flag, OTA over-rides
                    tempList[-1]["Stereo"] = 'stereo' in aplst
                    tempList[-1]["SAP"]    = 'SAP' in aplst
                    tempList[-1]["DVS"]    = 'dvs' in aplst
                
                # If SD has video propertie.... they are optional
                vplst = aProgram.get("videoProperties") 
                if vplst :
                    #logging.info('ch %s: video %s', aStation['Channel'], vplst)
                    tempList[-1]["HD"] = 'hdtv' in vplst  # True="HD", False="SD"  OTA over-rides
                    # DVR+ wants 'HD Level Unknown','480p','480p','720p','1080i','1080p'
                    # just one of these should be true - brute force seems easiest
                    if 'HD 480p'  in vplst : tempList[-1]["HDLevel"] = "480p"
                    if 'HD 720p'  in vplst : tempList[-1]["HDLevel"] = "720p"
                    if 'HD 1080i' in vplst : tempList[-1]["HDLevel"] = "1080i"
                    if 'HD 1080p' in vplst : tempList[-1]["HDLevel"] = "1080p"

                # content ratings - dont know if the SD ratings are compatible, i'm just
                # going to stuff whatever SD has into the DVR+ field
                # DVR+ seems to want: None,TV-G,TV-PG,TV-14,TV-Y works TV-MA does not
                
                rtlst = aProgram.get("ratings")  # Ratings properties are optional
                if rtlst :
                    if len(rtlst) > 1 :
                        logging.info('Ch %s: has %d ratings.', aStation['Channel'], len(rtlst) )
                    
                    # search for "USA Parental Rating" 
                    for rating in rtlst :
                        if rating['body'] == "USA Parental Rating" :
                            break
                    
                    # if found
                    if rating['body'] == "USA Parental Rating" :
                        #logging.info('ratings %s', rtlst)
                        ratingstr = rating.get('code')  # SD "body" : "USA Parental Rating"
                        tempList[-1]["TVRating"] = ratingstr
                    else :
                        #ratingstr = rtlst[0].get('code')  # otherwise take the first one?
                        tempList[-1]["TVRating"] = "None"     # fill with empty 
                        
                                        
        #logging.debug("LS:Build Part 2 Station.  New Airings = %d", len(tempList))
        

        # Part2 ========================================================================
        # Reqest SD program details for each program airing we just added
        # 
        # There are about 336(24 1-HR listing*14days) to 672(48 30-MIN listing*14days)
        # programs in the list (550 typical) for each day
        
        # request program details from Sched Direct for this channel in one big block
        # The SD call lets you request multiple ProgramID at one time
        # There is a note on the SD site to limit the request to 500 per call
                
        # Testing server seems to want less that 600 per call
        if (len(PgmIdList) < 600 ) :
            code, pdj = SDGetProgramDetails( PgmIdList  )
        else :
            # If the list is still too big, note it and make two calls to SD
            logging.info("LS:BLS: Chan %s PgmIdList > 600! Splitting into 2 groups.", aStation['Channel'] )
            logging.debug("LS:BLS: Chan %s  PgmIdList(%d) DupIdList(%d)", aStation['Channel'], \
                          len(PgmIdList), len(DupIdList) )
            
            # SD server was choking on lists > 600 program IDs with a 500 server error
            # so just get 2 lists and join the results together
            list1 = PgmIdList[0:575]
            code, pdj1 = SDGetProgramDetails( list1  )

            if code == 200 :
                list2 = PgmIdList[575:]
                code, pdj2 = SDGetProgramDetails( list2  )
                pdj = pdj1 + pdj2
                
            logging.debug("LS:BLS:  L1(%d) L2(%d) pdj1(%d) pdj2(%d) pdj(%d)", len(list1),len(list2),len(pdj1), len(pdj2), len(pdj) )
            
        if code != 200 : # some kind of bad error - clean up and exit
            MasterAiringList.clear()
            logging.warning("LS:BLS: Could not get program details. List not built! (%d)", code)
            BuildAbort = True   # Method to break out of Build process
            break
        
        # Now we cycle through our temporary list of airings
        for airing in tempList :
            
           # and find the matching programID in Prog Details Json pdj list we just requested
            for pd in pdj:
                if pd.get('programID')==airing['ProgramId'] : break
            
            #it should be in there since we built the list from teh ID's SD gave us...
                
            if pd.get('programID') != airing['ProgramId'] :
                logging.warning("LS:BLS: ProgramId (%s) not found!", airing['ProgramId'])
            
            #???The remaining keys must be looked up later from the Program ID's
                
            title = airing["Title"] = copy.copy( pd['titles'][0]['title120'] )
            
            # Make an enhanced program description
            # SH type programID's dont have descriptions
            # ToDo their could be an option to select 100 or 1000 
            # DVR+ can display up to 512 chars.  COPY_MAX_CHARS=512
            
            # How do we pick which description to use? d100 or d1000?
            # We have 3 things to consider
            # 1) do they want the Episode Title, S#E#, AirDate:?
            # 2) How long of a description can we add?
            # 3) do they want the cast list?
            # That seems to be the priority....
            # 1) Give them the Episode info if they asked it, then
            # 2) give them the longest descripton that will fit, staying under 512 chars
            # 3) then fit in as many full cast names as we can in 512 chars

            # Start the description with '(New)' for new episodes
            summary = "(New) " if (airing["AiringType"] == "New") else ""
                            
            #1 Build the Episode info string
            if PiGScfg['PrependEpisode'] :
                oAirDate = pd.get("originalAirDate","")
                
                #if oAirDate == "1999-09-05" :  # remove strange date from certain news broadcasts
                #    oAirDate = ""
                    
                if oAirDate : oAirDate = "Aired: " + oAirDate + ". "
                
                # not all listings have episode titles
                EpisodeTitle = pd.get("episodeTitle150","")
                if EpisodeTitle : EpisodeTitle = '"' + EpisodeTitle + '."  '   
                airing["EpisodeTitle"] = copy.copy(EpisodeTitle);
                # ET does nothing for DVR+ but there is a place for it so I will put it there 
                                
                # build a Season Episode S#E# string if we can
                seasonStr = ""
                md = pd.get("metadata")
                if (md) : #if metadata exists
                    if ( md[0].get("Gracenote") ) :  # only want gracenote meta
                        # sometimes there is 0,0 data which seems to be bogus b/c
                        # the same 0,0 occurs for multiple episodes
                        # epi title is usually missing and airdate seems like a default
                        
                        sn = md[0]["Gracenote"]['season']
                        en = md[0]["Gracenote"]['episode']
                        #seStr = "Season " + str(sn) + ", Episode " + str(en) + ". "

                        # Make a compact string so as not to use up too much description
                        seasonStr = "S" + str(sn) + ", E" + str(en) + ". "
                        
                        # if its an invalid season #, clear the Epi,SN,EN
                        if sn == 0 :
                            EpisodeTitle = oAirDate = seasonStr = ""
                
                summary += EpisodeTitle + seasonStr + oAirDate

            #2, handle description, also called 'copy'
                
            desc = "No summary available(PiGS)."  # default value, if we find no description
                                                  # this will give DVR+ something to display
            de = pd.get('descriptions')           # reference to the program description stuff
            
            # there can be a 1000 char descr, a 100 char descr or both, decide which to use

            d1000s = d100s = ""  # make sure strings are clear from prior iteration
            if de : # if there is a description section, in returned data
                d1000  = de.get('description1000')  # chk avail for long and short versions
                d100   = de.get('description100')
                
                # Then pull the strings for short and long if possible, or "" if not
                d1000s = d1000[0]['description'] if d1000 else ""
                d100s  =  d100[0]['description'] if d100  else ""
                    
            # Now we can decide which desc to use based on available space remaining in 512 limit
            #charsAvail = COPY_MAX_CHARS - len(summary)- 5 # 5 for end count below
            charsAvail = COPY_MAX_CHARS - len(summary)

            # now append the main description
            if len(d1000s) < charsAvail :  # if the long desc will fit, use it
                summary += d1000s
            elif len(d100s):  # else if there is a d100s, use that (even if its too long)
                #logging.debug("Using d100s CharsAvail(%d), d1000s(%d), d100s(%d)", charsAvail, len(d1000s), len(d100s) )
                summary += d100s  
            else :           #but... if there is no d100s, then use the long one anyway
                logging.info("Default d1000 CharsAvail(%d), d1000s(%d), d100s(%d)", charsAvail, len(d1000s), len(d100s) )
                summary += d1000s  # longer than 512 will just get truncated by DVR+
                    
            # now fill in any remaining space with cast info, if they asked for it
            
            # If they want the castlist, try and build one
            caststr = ""
            if PiGScfg['AppendCast'] :    # does user want one?
                castlist = pd.get('cast') # is there a castlist in the SD Data?
                if castlist :
                    caststr    = " Cast: "  # Label
                    
                    # adjust chars avail  less current len, less cast header
                    #charsAvail = COPY_MAX_CHARS - len(summary) - len(caststr) - 5 # 5 for end count below
                    charsAvail = COPY_MAX_CHARS - len(summary) - len(caststr)
                    
                    clen = len(caststr)          # len of just the label for testing below
                    
                    # fit as many full names as we can
                    for person in castlist :                        
                        actorname = person['name'] +", "  # get a name, append a ','
                        alen = len(actorname)
                        if alen > charsAvail :   # check if name will fit, 
                            break                # quit when out of space
                        caststr += actorname     # otherwise add name and continue
                        charsAvail -= alen       # account for name len
                    
                    if clen == len(caststr) :    # if we could not add any names, remove the label
                        caststr = ""
                    else :
                        #replace the trailing comma with a period, slice off last two chars, add a period 
                        caststr = caststr[:-2] + "."
            
            # Append the cast list if one was built
            summary += caststr

            # add the char (count) to the end - for testing only.
            #summary += '(' + str(len(summary)) + ')'
        
            slen = len(summary)
            if slen > COPY_MAX_CHARS :
                logging.warning("LS:BLS: Desc=%d %s %s", slen, airing["AiringTime"], airing["Title"])
                
            # summary was created for debugging so I could view the string easily
            airing["Copy"] = copy.copy( summary ) 
            
        # Part3 ========================================================================
        # Appends the temp program airings to the Master list, then sorts by date
        # https://docs.python.org/3/howto/sorting.html
        # https://www.geeksforgeeks.org/ways-sort-list-dictionaries-values-python-using-itemgetter/?ref=rp

        # add temp list to the master list
        MasterAiringList.extend(tempList)
 
        # sort the list by ascending airing list, compute sort time.  it is quick
        t1=datetime.utcnow()
        MasterAiringList.sort( key=itemgetter('AiringTime') )
        t2=datetime.utcnow()
        
        # Done with 1 station, report the completion, update progress        
        
        tdiff=t2-t1
        ts=str.format('{:1}.{:06}', tdiff.seconds,tdiff.microseconds)
        logging.debug("LS:BLS: %d airings addded. %d total items. Sort time: %ss", len(tempList),len(MasterAiringList), ts)
        #logging.info("LS:BLS: DupList %s", DupIdList)
        logging.info("LS:BLS: Added %s programs for channel %s, RfChan=%s (StationID=%d)", \
                        len(tempList), aStation['Channel'], aStation['ChannelLocation'], iSourceId)
        
        # compute build percent
        pct = aStation["Order"] / TotalChannels * 100
        BuildInfo['Schedule Build Info']['Build_Progress'] = str(int(pct)) + "%"
        
    # Build complete - wrap up
    # update schedule build time for Stale check later
    LastBuiltTime = datetime.utcnow()  # time of this request
    lt,lz = LocalTimeStr(LastBuiltTime, True)

    # if Build process was aborted due to SD error or Thread abort signal
    if BuildAbort :
        MasterAiringList.clear()
        PiGScfg["LastSchdBuild"] = "Build Aborted!"
        BuildInfo['Schedule Build Info']["Build_End_Time"] = "Build Aborted at " + lt
        logging.info("LS:BLS: Build Aborted at " + lt)
    elif len(MasterAiringList) > 0:
        logging.info("LS:BLS: Build Complete! %s total programs. Sched ends:%s", len(MasterAiringList), MasterAiringList[-1]['AiringTime'])
        logging.info("LS:BLS: Calculating MasterAiringList[] memory usage...")
        ListMBs = total_size(MasterAiringList)/1024/1024
        logging.info("LS:BLS: MasterAiringList[] size %4.1f MB", ListMBs )

        BuildInfo['Schedule Build Info']['Programs'] = len(MasterAiringList)
        BuildInfo['Schedule Build Info']['SchedMBs'] = str.format('{:.1f} MB', ListMBs)
        BuildInfo['Schedule Build Info']['SchdEnds'] = MasterAiringList[-1]['AiringTime']
        
        PiGScfg["LastSchdBuild"] = lt
        BuildInfo['Schedule Build Info']["Build_End_Time"] = lt
        logging.info("LS:BLS: Schedule Built at %s", lt)
    else :
        logging.warning("LS:BLS: Build Complete! %s total programs. Something Misconfigured?", len(MasterAiringList) )

    # compute build percent
    pct = 100
    BuildInfo['Schedule Build Info']['Build_Progress'] = str(int(pct)) + "%"
    logging.info("LS:BLS: Total Build time %s\n", datetime.utcnow()-BuildStartTime)

    #debug, print out first few items to verify sort, start date, etc
    limit=15
    logging.debug( 'LS:BLS: First %d items on schedule: ', limit )
    for airing in MasterAiringList :
        limit -=1  
        if limit>0 : 
            logging.debug("  %s %s %s", airing['AiringTime'], airing['Channel'], airing['Title'])
        else:
            logging.debug("  %s %s %s\n", airing['AiringTime'], airing['Channel'], airing['Title'])
            break

    return None



# Message type 2 - Return the a schedule when DVR+ request it  
# The http GET message looks like this:
# GET /TVlistings/v9/listings/linearschedule/20499/info?
#        apikey=v4edzwwtca4e847n72j8kesw&
#        locale=en-US&
#        duration=240&
#        inprogress=true&
#        startdate=2020-05-02T16:00:00Z&
#        format=json HTTP/1.1
#
# GetLinearSchedule() returns a DVR+ formatted json message of all listings that 
# *Start* 'in' the DVR+ requested time window :
#    (startdate) through (startdate+duration) 
# If the flag inprogress is set to true, then GLS will also return listings that
#  Start *before* the request window, but end *in* the request window
#
# debug if ChFilter is a string like '15.1' then only the sched for that channel is returned
#  this is not implemented right now

# Maintain a list of the DVR IP's that are making requests.  Save their ListIndex & StartTimes. 
# DVR IP will be the key, the value is a list with the List Index, and the tFirstRequest

DictOfDVRs = {}
# will look like this as various dvr's make requests to PiGS
# { 'DVR1-IP' : [ListIndex1, StartTime1, LastUpdateStr1, RequestCounter1],
#   'DVR2-IP' : [ListIndex2, StartTime2, LastUpdateStr2, RequestCounter2],
#   'DVR3-IP' : [ListIndex3, StartTime3, LastUpdateStr3, RequestCounter3],
#   ....
# }

# Index offsets to List Items
STATE_LI = 0    # List Index
STATE_ST = 1    # Start Time
STATE_LU = 2    # Last Update
STATE_RC = 3    # Request Counter

# For Status display in a browser
def GetDVRStates() :
    global DictOfDVRs
    
    # cant json datatime objects so convert list item [1] to str
    # and make a more pretty json object
    dj = {}
    for k in DictOfDVRs :
        dj[k] = {}
        dj[k]['ListIndex'] = DictOfDVRs[k][STATE_LI]
        dj[k]['RequstNum'] = DictOfDVRs[k][STATE_RC]
        
        ts,tz = LocalTimeStr( DictOfDVRs[k][STATE_ST],True )
        dj[k]['FirstRqst'] = ts
        ts,tz = LocalTimeStr( DictOfDVRs[k][STATE_LU],True )
        dj[k]['Last_Rqst'] = ts
        
        tdiff = DictOfDVRs[k][STATE_LU]-DictOfDVRs[k][STATE_ST]
        ts=str.format('{:02}:{:02}s', tdiff.seconds//60, tdiff.seconds%60)
        dj[k]['__Elapsed'] = ts
        
    return dj

def GetLinearSchedule(PiGScfg, qs, serviceID, DVR_IP, tzlist, ChFilter=None):

    global DictOfDVRs

    theTime=datetime.utcnow()  # time of this request

    # parse the GET params
    dur     = int(qs["duration"][0])      # duration in minutes
    bInProg = qs["inprogress"][0]=='true'
    rsts    = qs['startdate'][0]          # request start time string
    
    # Figure out DVR+ request window start and end times in Zulu
    rst     = parser.parse( rsts )
    ret     = rst + timedelta(minutes=dur)

    # Fixing Frank70's multiple dvr's at the same time bug.
    # Cache a list of DVRs making schedule requests to this server
    # keep track of their list index and start time 
    if DVR_IP not in DictOfDVRs :
        DictOfDVRs[DVR_IP] = [0,0,0,0]     # if key does not exist, create it and initialize
        
    # Get the ListIndex for this DVR, ListIndex will be cleared below if this is the first request
    ListIndex = DictOfDVRs[DVR_IP][STATE_LI]
    
    logging.info("LS:GLS DVR %s State[%d,%s]", DVR_IP, DictOfDVRs[DVR_IP][STATE_LI], DictOfDVRs[DVR_IP][STATE_ST] )

    # Update the Response Header
    # Most of the stuff in the header is ignored by DVR+
    #%f is 6-digit microsecs.  need mS?  setting to '000' 
    lsr["LinearScheduleResult"]["TimeStamp"] = theTime.strftime("%Y-%m-%dT%H:%M:%S.000Z")
    lsr["LinearScheduleResult"]["Schedule"]["ServiceId"]= serviceID  # always 99999

    lsr["LinearScheduleResult"]["Schedule"]["TimeZones"] = copy.deepcopy(tzlist)
    
    lsr["LinearScheduleResult"]["Schedule"]["StartDate"] = qs['startdate'][0]
    lsr["LinearScheduleResult"]["Schedule"]["Duration"]  = dur
   
    #https://www.w3schools.com/python/python_dictionaries.asp
    # get reference to the airings list in result dict object
    
    alist = lsr["LinearScheduleResult"]["Schedule"]["Airings"]
    alist.clear()  # python 3.0, clear the list
    
    # start at the beginning of the list (ListIndex=0) on the first DVR+ request (bInProg=true)
    #  on subsequent requests leave ListIndex alone to start where we left off for speed.
    # On the first request there could be shows on the schedule grid that are
    #  inProgress.  like if we are viewing at 730pm, a show from 7pm could be ongoing.
    # DVR+ needs to know about this and we can use it to print start time, reset search index, etc
    if bInProg :
        ListIndex = 0
        
        DictOfDVRs[DVR_IP][STATE_ST] = datetime.utcnow()          #tFirstRequest = datetime.utcnow()
        DictOfDVRs[DVR_IP][STATE_RC] = 0
    
        #logging.info("LS:GLS First Request at UTC: %s", tFirstRequest.strftime("%Y-%m-%dT%H:%M:%SZ"))
        tss, tzz = LocalTimeStr( DictOfDVRs[DVR_IP][STATE_ST], showSeconds=True)
        logging.info("LS:GLS First Request at: %s", tss )

    # error condition - if PiGS was restarted somehow during a DVRs schedule request,
    # The dvr will continue on its Merry way, but the Start Time would be unitialized at 0
    
    if DictOfDVRs[DVR_IP][STATE_ST] == 0 :
        DictOfDVRs[DVR_IP][STATE_ST] = datetime.utcnow()          #tFirstRequest = datetime.utcnow()
        logging.warning("LS:GLS Unexpected unitialized Start Time.  Setting to UTC now!")

    # Get the start time for this particular DVR, time could get reset just above
    tFirstRequest = DictOfDVRs[DVR_IP][STATE_ST]

    # Total items in the AiringList
    listlen = len(MasterAiringList)

    # if the list is empty, its wierd, but not the end of the world.
    # we will just return an empty list and DVR+ will be happy
    if not listlen  :
        logging.warning("LS:GLS: MasterAiringList has not been built!!")

    # As a DVR or DVR's make schedule requests, we just leave the ListIndex at its last position
    # because the DVR is going to make the next request right there anyway.  minimizes seaching.
    
    # Dont have to test for accessing beyond the end of the list because the search 
    # method below backs up at least 1 listing before accessing MasterAiringList[]
    
    if (listlen>0) : # and (ListIndex<listlen) : 
        
        # Test for a request time before the listIndex.
        # In case a DVR does a retry we may need to back up the search index pointer ListIndex.
        # search backwards until I find a listing before the rst, or I get to the beginning of the list.
        # if this 'Backup' linear search method got too slow, some type of binary search could be implemented

        lis = ListIndex             # remember where we started for logging info
        
        while (ListIndex > 0 ) :    # search, but not before the start of the list.

            #  back up 1 listing and check listing time.
            ListIndex -= 1    # pre-decrement to ensure not accessing past the end
            
            #get program airing start time STRING like "2020-03-08T08:00:00Z"
            pss = MasterAiringList[ListIndex]['AiringTime']
            
            # convert to a TIME - prog start time pst
            pst = parser.parse( pss )
            
            # if the Program Start time pst is before the requested start time rst, were are done
            if (pst < rst) : break      # out of the loop
            
        
        # backup of 1 is expected
        if (lis - ListIndex) > 1 :   
            logging.info('LS:GLS:Request before current ListIndex(%d). Rollback to ListIndex(%d).', lis, ListIndex)
 
    logging.info('LS:GLS:Searching for %s. InProgessFlag=%d...', rsts,bInProg)

    ListingCnt = 0   # Count how many we send in each request, just for Info
    
    while (ListIndex < listlen) :   # dont search past end of list, could be 0 items

        #get program airing start time string like "2020-03-08T08:00:00Z"
        pss = MasterAiringList[ListIndex]['AiringTime']
        #prog start time pst
        pst = parser.parse( pss )
        #prog duration int
        pdur= MasterAiringList[ListIndex]['Duration']
        #prog end time pet
        pet = pst + timedelta(minutes=pdur)

        # if the Program Start time pst is beyond requested end time ret, were are done
        if (pst >= ret) : break      # out of the loop 

        # test if the start of this program is in the req window (pst>=rst)
        # or
        # if asking for an inprogress shows, it 'starts before the window' and 'ends in the window'
        bInWindow = (pst>=rst) or \
                     bInProg and (pst<rst) and (pet > rst)
  
        # debug if being asked to filter to 1 channel, 
        #if ChFilter :
        #    bInWindow = bInWindow and ChFilter == MasterAiringList[ListIndex]['channel']
        #if bInWindow : print("li bInWin ", ListIndex, bInWindow)
  
        # this compiles list of references to the Master list, so dont delete Master list!
        
        if (bInWindow) :
            
            ListingCnt += 1
            
            alist.append( MasterAiringList[ListIndex] )
            
            # could deep copy the data, but we should not need to 
            # create a new Channel structure, recursive dict copy
            #alist.append( copy.deepcopy(MasterAiringList[ListIndex]) )
            
            #debug print a list of New shows
#            bnew = MasterAiringList[ListIndex]['AiringType'] == 'New'
#            if bnew :
#                a = MasterAiringList[ListIndex]
#                airtimez = parser.parse( a['AiringTime'] )
#                airtimecst = airtimez + timedelta(minutes=-300)            
#                print("New", a['Channel'], airtimecst, a['Title'])
                
        ListIndex += 1;  # move on to the next show in MasterAiringList
        
    #end of while (ListIndex < listlen) :

    t2=datetime.utcnow()
    tdiff = t2-tFirstRequest
    #ts=str.format('{:02}:{:02}.{:06}s', tdiff.seconds//60, tdiff.seconds%60,tdiff.microseconds)
    ts=str.format('{:02}:{:02}s', tdiff.seconds//60, tdiff.seconds%60)
 
    # Done serving this request, Save the ListIndex for this DVR for it's next call
    # Update this DVR's State

    DictOfDVRs[DVR_IP][STATE_RC] += 1         # count number of requests
    DictOfDVRs[DVR_IP][STATE_LI] = ListIndex  # 
    DictOfDVRs[DVR_IP][STATE_LU] = t2         # Last Update Time in Local Time
    
    rqCnt = DictOfDVRs[DVR_IP][STATE_RC]

    logging.info('LS:GLS:Request complete! %d Listings. LI=%d RequestCount=%d, Elapsed= %s\n', ListingCnt, ListIndex, rqCnt, ts)


    # If we have no more data to send to the DVR we have to send it an empty list
    # on the final request the DVR+ gets/needs an empty list airings=[]
    # but with SD we get more than 2 weeks of data, so DVR+ just quits asking after 14days (or 90 requests
    # I see this if the DaysLimit is set to something less than 16 in the config file
    # so this message doesnt always get printed - it depends on the data
    # It's been empirically determined that the DVR asks for 90 4 hour blocks and then stops asking
    
    if len(alist) == 0 :
        #tdiff = t2-tFirstRequest
        #ts=str.format('{:02}:{:02}.{:06}s', tdiff.seconds//60, tdiff.seconds%60,tdiff.microseconds)
        #logging.info('LS:GLS:That appears to be the final request for (%s)! Total time =%s\n', DVR_IP, ts)
        logging.info('LS:GLS:PiGS is out of data. Sending empty response to DVR (%s)!\n', DVR_IP)
 
    if rqCnt == 90:
        logging.info('LS:GLS:That appears to be the final request for DVR (%s)! Elapsed Time %s\n', DVR_IP, ts)
        
    #convert Python dict to json and return
    js = json.dumps(lsr)    
    
    return js

