# pc.py
#
# Copyright June 2020, Timothy Goihl
#
# uses cm.py
#
# A module to test the quality and data integrity of the Channel Master Guide Data
#
# Also run a SimpleHttpServer file server to test if PiGS might run on
# other platforms like Windows a Mac, or Linux machine
#

import json

ZIP_TO_SCAN = '53593'       # PostalCode to scan.  in single quotes

# List of channels to analyze.
# Only these channels will be scanned for broken Airings
chanlist = ["3.1", "3.2", "3.3", "3.4", "3.5", "15.1",'15.2', '15.3', '15.4', '21.1','21.2', \
           '21.3','21.4','27.1','27.2','27.3','27.4','27.5','47.1','47.2','47.3','47.4',\
           '57.1','57.2','57.3','57.4','57.5','57.6','57.7','57.8','57.9' ]
chanlist = None  # comment out this line to scan every channel in the CM Service Code

import logging
LOG_LEVEL   = logging.INFO # logging.INFO or logging.DEBUG

# To use this test program, open pc.py in your python editor and click run.
# it should first attempt to read data from the CM/Rovi servers and print some results
# That tests outgoing http commands
#
# Next it will attempt to start a server... if run from the programming environment
# it defaults to port SERVER_PORT=8000 instead of 80 for webservers. See constant below.
# on the Raspberry Pi platform, you could NOT set SERVER_PORT to 80 in the programming GUI
# because it would bail with an error that port 80 was 'protected.'
#
# To start on port 80 I had use a Pi Terminal/console window and type
#     sudo python pc.py 80
#   or 
#     sudo python3.7 pc.py 80
#
# pc.py will print a message as to the port it is using.  0.0.0.0 is ok, it means all IP's
#
# maybe the MS Windows Python will let you change the constant SERVER_PORT to 80 and allow
#   it run it from the UI?
# If not, figure out how to do the sudo type thing above in windows.
# Basically you need to see a message like this without an error msg right after it:
#
#   INFO:root:HTTPServer binding to IP=0.0.0.0  Port=80.
#
# To test the server side, open a browser from another PC or phone and type the IP of your PC
# and the name of this file -
# If the server is working the browser should list the contents of this file.
#
# 192.168.1.16/pc.py         for port 80
#   or
# 192.168.1.16:8000/pc.py    for port 8000
#
# If the file displays, your PC is a server!!

SERVER_HOST = '0.0.0.0'    #'localhost' did not work for outside requests
SERVER_PORT = 8000         # to run with port 80 cmd line: "sudo pyton3.7 pigs.py 80"

#Programming references
#https://pynative.com/python-post-json-using-requests-library/
#Chunked encoding
#https://requests.readthedocs.io/en/master/user/advanced/#chunk-encoded-requests
#https://requests.readthedocs.io/en/master/user/quickstart/#json-response-content
#https://requests.readthedocs.io/en/master/user/advanced/#request-and-response-objects
#https://requests.readthedocs.io/en/master/user/quickstart/
#>>> payload = {'key1': 'value1', 'key2': 'value2'}
#>>> r = requests.get('https://httpbin.org/get', params=payload)
#https://dateutil.readthedocs.io/en/stable/

from http.server import HTTPServer, BaseHTTPRequestHandler, SimpleHTTPRequestHandler
from io import BytesIO
from urllib.parse import parse_qs, urlparse

# since we cant use the epg name to get to cm (b/c of dns redirect),
# we will go more directly using the Alt DNS cname DNS response 0x25e4 A epg.channelmastertv.com
#CNAME proxy-elb-1800886808.us-west-2.elb.amazonaws.com A 34.212.152.253 A 54.148.61.14

CM_EPG_SERVER     = "epg.channelmastertv.com"
ALT_CM_EPG_SERVER = "http://proxy-elb-1800886808.us-west-2.elb.amazonaws.com/"

# CM API Links
#GET /TVlistings/v9/listings/services/postalcode/53593/info?
#    apikey=v4edzwwtca4e847n72j8kesw&locale=en-US&countrycode=US&msoid=306430&format=json
CM_SERVICE_URL = "TVlistings/v9/listings/services/postalcode/"

#GET /TVlistings/v9/listings/servicedetails/serviceid/20499/info?
#    apikey=v4edzwwtca4e847n72j8kesw&locale=en-US&includechannelimages=true&
#imagehorizontalresolution=92&imageverticalresolution=36&format=json
CM_SERVICEDETAILS_URL = "TVlistings/v9/listings/servicedetails/serviceid/"

#GET /TVlistings/v9/listings/linearschedule/20499/info?
#    apikey=v4edzwwtca4e847n72j8kesw&locale=en-US&
#    duration=240&inprogress=true&startdate=2020-05-02T16:00:00Z&format=json
CM_LINEARSCHED_URL = "TVlistings/v9/listings/linearschedule/"

import json
import requests
import logging
import datetime
import sys

from cm import CMGetServicesFromPostalCode
from cm import CMGetLinearSchedule
from cm import CMGetServiceCode
from cm import CMGetServices

# to re-import in interactive mode...
#https://switowski.com/blog/ipython-autoreload
#https://stackoverflow.com/questions/4385839/reimporting-a-single-function-in-python
#import sys
#from importlib import reload
#reload(sys.modules['cm'])
#reimport your function


##################################################################################
# TestRoviListingData - look for irregularities

DayCounts = []   # list of daycount

from datetime import timedelta, datetime, date
from dateutil import parser
from copy import copy

# Local Time convert
#http://robsraspberrypi.blogspot.com/2016/02/raspberry-pi-all-things-date-and-time.html
# To get the Pi's current time zone UTC offset do the following
# import time
# print "The UTC offset is",time.altzone/3600,"Hours"
# returns +5, meaning UTC is local time +5 hours
#https://docs.python.org/3/library/time.html
#tz = timedelta(seconds=time.timezone if (time.localtime().tm_isdst == 0) else time.altzone)

import time
def LocalTimeStr(at) :  # doing my own local time convesion for my 'unaware' times
    
    if time.daylight :  # determine which offset to use based on daylight flag (0 or 1)
        tzhrs = -time.altzone/3600
    else :
        tzhrs = -time.timezone/3600
        
    # Get the timezone name
    tzstr = "(" + time.tzname[time.daylight] + ")"
    
    localTime = at + timedelta(hours=tzhrs)
    return(localTime.strftime("%Y-%m-%d %H:%M") + tzstr )

# ZuluFilter format 2020-05-02T16:00:00Z
# chfilter eg "3.1"
# maxdays=0 requests max avail
# start date format 2020-05-02  (empty starts 'today')
# DisplayFilter display all listings for a list channel
# chlist   list of channels - only count/print for Channels in the list, None Does all


def TestRoviListingData(postalcode, maxdays=20, startdate ="", DisplayFilter = [], chlist=[], ZuluFilter ="" ) :
    
    global DayCounts;
    
    # Get the service code sc, like 20499 for zip 53593
    # sc is a string so it is easy to use to build the next urls
    code, sc = CMGetServiceCode(postalcode)
    if code != 200 : return code
    
    # Test the Services structure for Channel match VCN, bad service codes, etc
    #print(f'Checking ServiceDetail Lineup zip({postalcode}), ServiceId({sc})...')
    logging.info('Checking ServiceDetail Lineup zip(%s), ServiceId(%s)', postalcode, sc)
    codess, sdjson = CMGetServices(sc)
    if codess == 200 :
        if len(sdjson) == 0 :
            logging.info("Service Detail list is has no channels!")
        
        chlineup = sdjson['ServiceDetailsResult']['ChannelLineup']
        if chlineup['ServiceId'] != int(sc) :
            logging.info("Service Detail Header ServiceId does not match!")
            
        svchlist = chlineup['Channels']
        for ch in svchlist :
            
            if (ch['ServiceType']=='Digital') and \
                        (ch['Channel'] != ch['VirtualChannelNumber']) :
                #print(f"Channel({ch['Channel']:4}) VChannel({ch['VirtualChannelNumber']:4}) Error!")
                logging.info("Channel(%4s) VirtualChannel(%4s) Error!", ch['Channel'], ch['VirtualChannelNumber'])
                
            # The 'SourceId' key in each channel here should match the ch/SourceId in the listings

        zones = sdjson['ServiceDetailsResult']['ChannelLineup']['TimeZones']
        logging.info("CM TZ Info %s", zones)

    # now test main schedule database
    # things to count
    daycount = {"adate": "2020-00-00T00:00:00Z", "counts":{} }
    
    # this is used to reset counts object
    # NoTitleKey - the json key is missing from the airing
    # EmptyTitle - the key exists, but the string is empty or null
    # NoCopyKey -  the json key is missing from the airing for program description
    # EmptyCopy -  the key exists, but the string is empty or null for program description
    #
    # BadSvcId   - Service ID in a listing does not match header service IF
    # Could verify that the sourceID/stationID matches service/channel info 
    
    counter = { "TotAirings":0, "NoTitleKey": 0,"EmptyTitle": 0, "NoCopyKey":0, "EmptyCopy": 0, \
                "BadSvcId": 0 #"HasEpisodeTitle": 0
                }
                #"BadStation", 0, "DurIs0": 0, }
  
    dur = 240  # 4 hr blocks just like DVR+
    inprogress = "true"   # for the first
    
    if maxdays == 0 : maxdays = 21  # set to 3 weeks... 'never' that much available

    theTime=datetime.utcnow()  # time of this request, default to now
    
    if startdate :             # if user supplied a starting date, use that
        theTime = parser.parse(startdate, ignoretz=True)

    #https://docs.python.org/2/library/datetime.html,   truncate to start at 0Z
    theTime = theTime.replace(hour=0, minute=0, second=0, microsecond=0)
    endTime = theTime + timedelta(days=maxdays)
    
    # a string is needed for the url
    startTime = theTime.strftime("%Y-%m-%dT%H:%M:%SZ")
    
    dayroll = theTime.day  # watch for the day to roll to next day for subtotal counts
    counts = copy(counter) # start with zeros

    ChTitleList = []
    ChDescList = []
    ChDescDictCnts = {}
    
    logging.info("\nScanning the CM/Rovi database for errors. PostalCode(%s)", postalcode)

    while theTime < endTime : # test for empty list, this does loops of dur=240 minutes 
         
        # make a string for the command, save time start of this request, bumps by 4 hrs each request
        tstr = theTime.strftime("%Y-%m-%dT%H:%M:%SZ")
        
        #The rovi V9 LinearSchedule api has a 'bug' that returns airings that begin at the end of the 
        #request window, and then again at the beginning of the next window, so those airings get sent 2x.
        #So, if the prog starts exactly at the end time, skip it so its not double counted
        #The same program will be returned at the start of the next window.
        #Rovi has test logic something like
        #         if (AiringTime >= ReqStart && AiringTime <= ReqStart+Dur) SendIt!
        # but it should be 
        #         if (AiringTime >= ReqStart && AiringTime <  ReqStart+Dur) SendIt!
        
        endwin = theTime + timedelta(minutes=dur)

        # make the schedule request to CM Alternate server
        code, ls = CMGetLinearSchedule(sc, tstr, dur, inprogress)
        if code != 200 :
            logging.warning("PC.py  CMGetLinearSchedule error(%d) at %s", code, tstr )
            return code
        
        # get service ID from the returned schedule
        headerSID = ls['LinearScheduleResult']['Schedule']['ServiceId']
        
        # reference to list of all ariings in the schedule for brevity
        airinglist = ls['LinearScheduleResult']['Schedule']['Airings']
        
        if len(airinglist) == 0 :  # if no more data available, break out
            #print("End of list!")
            break
        
        inprogress = "false"  # flag only true on first request, reset 
        
        # check each airing 
        for a in airinglist :
            
            # get airing time, treat as unaware UTC for end window compare 
            at = parser.parse( a['AiringTime'], ignoretz=True )

            # check for V9 API bug.  exact compare to end of window
            if at == endwin :
                #print("end window, skipping")
                continue
            
            # only record one specific date and time.  
            if ZuluFilter :
                if ZuluFilter != a['AiringTime'] :
                    continue

            # only record one specific channel, like "3.1"
#            if DisplayFilter :
#                if a['Channel'] in DisplayFilter :
#                    print("ChFilter: ", a['AiringTime'], LocalTimeStr(at), a['ProgramId'], a['Channel'], a.get('Title'))
#                continue
                
            # only inspect/count my local channels
            if chlist :
                if a['Channel'] not in chlist :
                    continue;
                
            # Here is the start of the error checking / data validation
            #Do CM/Rovi listings have episode titles?  Yes, they do.  not doing this test
            #if not a.has_key('Title') :
            #if 'EpisodeTitle' in a :
            #    if len( a['EpisodeTitle'] ) > 0 :
            #        counts['HasEpisodeTitle'] +=1
            
            counts['TotAirings'] +=1   # count total airings for each day
            
            if 'Title' not in a :    # check if key exists in this airing
                counts['NoTitleKey'] +=1
                #print("No Title key: ", a['AiringTime'], LocalTimeStr(at), a['ProgramId'], a['Channel'],sep='\t')
                
                if a['Channel'] not in ChTitleList : ChTitleList.append(a['Channel'])
                
                if DisplayFilter :
                    if a['Channel'] in DisplayFilter :
                        print("ChFilter: ", a['AiringTime'], LocalTimeStr(at), a['ProgramId'], a['Channel'], a.get('Title'))
                #set logging to debug for details on missing channels 
                logging.debug("NoTitleKey: %s %s %8d %8s %4s", a['AiringTime'], LocalTimeStr(at), a['SourceId'], a['ProgramId'], a['Channel'])
                pass
            else :                      
                if len(a['Title']) == 0 :
                    counts['EmptyTitle'] +=1

            if 'Copy' not in a :
                counts['NoCopyKey'] +=1
                tstr = LocalTimeStr(at)
                
                if a['Channel'] not in  ChDescList : ChDescList.append(a['Channel'])
                
                cnt = ChDescDictCnts.get( a['Channel'], 0)
                cnt += 1
                ChDescDictCnts[ a['Channel'] ] = cnt

                #print("No Copy key: ", a['AiringTime'], a['ProgramId'], a['Channel'])
                #tstr = localTime.strftime("%Y-%m-%d %H:%MC")
#                tstr = localTime.strftime("%Y-%m-%dT%H:%M:%SC")
#                print("no key Copy: ", tstr, a['ProgramId'], a['Channel'])
                logging.debug("NoCopyKey: %s %s %8d %8s %4s", a['AiringTime'], LocalTimeStr(at), a['SourceId'], a['ProgramId'], a['Channel'])
                
            else :
                if not a['Copy'] :  # key exists, does it have content?
                    counts['EmptyCopy'] +=1
            
            #if a['Duration']==0 :  # never observed... not checking for this one
            #    counts['DurIs0'] +=1
                
            if a['ServiceId'] != headerSID :   # serviceID in header must match in Airing
                counts['BadSvcId'] +=1

            # could get the ServiceDetails, build a list of SourceIds[]
            # then do a if a['SourceId'] in SourceIdList 
            #a['SourceId'] must match the service id in Services Call
        
        # bump to next 4 hr block request
        theTime = theTime + timedelta(minutes=dur)

        #the day has just rolled over, save counts & reset
        if theTime.day != dayroll : 
            
            # add daycount object to the list 
            DayCounts.append( copy(daycount) ) # new object
            
            # Save the date and current counts in list
            DayCounts[-1]['counts'] = copy(counts)
            DayCounts[-1]['adate']  = copy(startTime)  # Record start time of this block
            
            # print the counts as we go
            logging.info('%s %s', startTime, counts)

            # save start time string
            startTime = theTime.strftime("%Y-%m-%dT%H:%M:%SZ")
            dayroll   = theTime.day      # save compare day
            counts    = copy(counter)    # reset counts
            
    # At this point there could be a partial day of airings.  Print it if there is.
    if counts['TotAirings'] > 0 :
        
        # add daycount object to the list 
        DayCounts.append( copy(daycount) ) # new object
        
        # Save the date and current counts in list
        DayCounts[-1]['counts'] = copy(counts)
        DayCounts[-1]['adate']  = copy(startTime)  # Record start time of this block
        
        # print the counts as we go
        logging.info('%s %s', startTime, counts)
        
             
    # When done, create one more entry that is the total for all these days
    totals = copy(counter) # start with zeros

    for dd in DayCounts : # [{"adate": "", "counts":""}]
        for key in dd['counts'] :
           totals[key] += dd['counts'][key]
    
    # Add it to the list, but give special name instead of a 'Date'
    DayCounts.append( copy(daycount) ) # new object
    DayCounts[-1]['counts'] = copy(totals)
    DayCounts[-1]['adate'] =  "            Totals>>"    
    logging.info('%s %s', DayCounts[-1]['adate'], DayCounts[-1]['counts'])

    print("List of missing titles:", ChTitleList)
    
    ChDescList.sort()
    print("List of missing Description Keys:", ChDescList)
    #print("List of missing Description Keys:")
    
    #for ccc in ChDescList :
        #print( ccc, ChDescDictCnts[ccc] )
    #print("List of missing Description Cnts:", sorted(ChDescDictCnts) )
#    scnts = sorted(ChDescDictCnts.keys())
    
    #print( sorted( ChDescDictCnts.items(), key = lambda kv:(kv[1], kv[0]) ) )
     
#    js = json.dumps(ChDescDictCnts, indent=2)
#    print("List of missing Desc Cnts:", js )
    
#    js = json.dumps(scnts, indent=2)
#    print("List of missing Desc Cnts:", js )
#
    return 200


###########################################################################################
#  Main code starts here
#  Get schedule data from ChannelMaster / Rovi servers and count how many bad listings there are


# initialize logging before making calls to any modules that use logging or weird(nothing) things happen
logging.basicConfig(filename="cmdata.log", level=LOG_LEVEL)  #.DEBUG will print more internal messages

console = logging.StreamHandler()  # this allows logging to console while logging to file also
console.setLevel(LOG_LEVEL)
logging.getLogger("").addHandler(console)

logging.info("Starting CM/Rovi Data test at %s", LocalTimeStr( datetime.utcnow() ) )

# list of channels to display more details on, not well thought out
DispList = ["3.1", "15.1", "27.1", "47.1"]
DispList = None


# Process command line arg to get port number
#if len(sys.argv) > 1:
#   logging.info("SYS args name=%s Port=%s", sys.argv[0], sys.argv[1])
#   port = int(sys.argv[1])
#else :
#   port = SERVER_PORT
#   logging.info("Using Port = %d", port )


# Get the zip code from the command line

if len(sys.argv) < 3 :
   logging.warning("Sys ArgV Count: %d", len(sys.argv) )
   logging.info("SYS args name=%s Port= Zip=", sys.argv[0] ) #, sys.argv[1], sys.argv[2])
   logging.warning("Bad command line. Usage:   python3 pc.py    <port> <zipcode>")
   logging.warning(" Webserver test port 8080  python3 pc.py     8080  <zipcode>")
   logging.warning(" Webserver test port 80    sudo python3 pc.py 80   <zipcode>")
   logging.warning(" No webserver test         python3 pc.py       0   <zipcode>")
#   exit(0)
else :
   logging.info("SYS args name=%s Port=%s Zip=%s", sys.argv[0], sys.argv[1], sys.argv[2])
   port = int(sys.argv[1])
   zipcode  = sys.argv[2]

if len(sys.argv) == 1 :
   port = 0
   zipcode  = '53593'
   logging.info("Incomplete command line, defaulting to 0, 53593")
   

code = TestRoviListingData(zipcode, maxdays=22, startdate="", chlist=chanlist, \
                                    ZuluFilter="", DisplayFilter=DispList)
#cnts = TestRoviListingData(ZIP_TO_SCAN, maxdays=0, startdate="2020-06-01", chlist=chanlist, \
#                                   ZuluFilter="", chfilter="")

#https://docs.python.org/3/reference/lexical_analysis.html#f-strings
if code == 200 :
    logging.info("\n==== Summary ====")
    dc = DayCounts[-1]['counts'] # ref to last item in the list
    logging.info("%s - %s (%d Days)", DayCounts[0]['adate'], DayCounts[-2]['adate'], len(DayCounts)-1)
    logging.info('Ch Filter: %s', chanlist)
    logging.info("Total Airings : %d", dc['TotAirings'] )

    for key in DayCounts[-1]['counts'] : # [{"adate": "", "counts":""}]
        if key == 'TotAirings' : continue  # dont display % for the total, it is 100%
        pct = dc[key]/dc['TotAirings']*100
    #    print(f'{key:10} {pct:6.1%} {dc[key]:4}' )
        logging.info("%10s %6.1f%%  %d" ,key, pct, dc[key])
else :
        logging.info("There was an error scanning the data")
        
###########################################################################################
# When that is done start a simple file server for testing different platforms
# If HTTPServer can server files at port 80, the platform can probably be a DVR+ server.

if port > 0 :
    logging.info("Starting Server...")
    logging.info('HTTPServer binding to IP=%s  Port=%d.',SERVER_HOST, port)

    # create the file server object
    try:
        httpd = HTTPServer((SERVER_HOST, port), SimpleHTTPRequestHandler)
    #except:
    except Exception as ex:
        template = "An exception of type {0} occurred. Arguments:\n{1!r}"
        message = template.format(type(ex).__name__, ex.args)    
        logging.warning("Error Starting server: %s", message )
        exit(0)
         
    logging.info("Waiting for requests!\n")

    # and start it... this call does not end, allow for clean exit from keyboard, Hit Ctrl-C
    try:
        httpd.serve_forever()
    except KeyboardInterrupt:
        pass

    httpd.server_close()

logging.info("Exiting... Done!")

