# pigs.py
#
# PiGS Channel Master CM-7500 DVR+ Guide Server Replacement
#
#---------------------------------------------------------------------------------------
# Copyright 2020, Timothy Goihl
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# See GNU-GPL-V3.txt License File
#
#---------------------------------------------------------------------------------------
#
# Files in the PiGS Guide Server project
#  pigs.py The main application / server
#  ls.py   PiGS LinearSchedule functions
#  sv.py   PiGS Service Details functions
#  cm.py   Channel Master / Rovi APIs (access CM/Rovi servers) to get image data
#  sd.py   Schedules Direct.org API calls
#  ut.py   Utility functions that may be used by more than one module
#  pigsdns.py  DNS server for DVR+ added in v1.02a
#  index.htm - configuration UI for pigs static IP e.g. //192.168.1.200
#  plus 5 image files
#  pc.py Test functions to CM/Rovi servers, also a Simple file server.
#     pc.py can be used to see if pigs.py will run on another hardware platorm.
#     pc.py needs cm.py to run, but that is it.
#     see instructions at the top of pc.py
# pigs.service   a file that can be used to start pigs at boot (ie Auto-Run)
# pigs.desktop   a file that can be placed on the Pi desktop to manually start PiGS
# GNU-GPL-V3.txt License File
# ReleaseNotes.txt - verbose history of version changes 
#
#---------------------------------------------------------------------------------------
# to run pigs.py from a Terminal window, type:
# sudo python3   pigs.py 80    
# Ctrl-C to exit/stop server
#
# Companion Spreadsheet with install and development Notes.
# https://docs.google.com/spreadsheets/d/1fg2dEN390CaZ0fP0DSSJjdO8EyvbOL_oaIsbLogmE8I/edit?usp=sharing
#---------------------------------------------------------------------------------------

PIGS_VERSION = "V1.12 2020-Oct-05"

# See ReleaseNotes.txt for detailed version history
#
# ToDo
# use import systemd.daemon to setup Auto-Start for user?
#
# V1.12 2020-Oct-05
#    Change LocalTimeStr() to work cross platform better.
#       Some display CDT, others Central Daylight Time was messing up DNS dipslay on /pigstatus
#    Fix dict init bug in ls.py lines 194-196
# V1.11 2020-Oct-03
#    Defaulted PinkPigPos to 0 for new installs.
#    Increased log file size to 150K before rollover.  Not quite enough look-back to solve issues
#    Add try/except in Builder calls to handle build exceptions, SD errors 
#    Add a 30 minute timer to retry the Service build if the Service build fails at boot
#    Fixes in HTML HEAD and unsupported commands
#    Added a DNS server to PiGS.  External DNS servers no longer needed.
#      sudo sysemctl disable dnsmasq, sudo sysemctl stop dnsmasq
#     /pigstatus Displays "DNS Enabled": "True, Last Request at 15:58:54(CDT)",
#    New Config Keys:
#      "IconType"     : 1,     # 0=No Icons, 1=Sched Direct, 2=CM/Rovi Icons, 3=Future???
#      "UseCMImages"  : key removed.  IconType takes its place
#      "PigsDoesDNS"  : True,  # PiGS acts as a DNS Server for DVR+.  Forward requests for msft,google,amzn, etc
#                                Defaults to False for upgrades so as not break an installation
#      "PigsDelayDNS" : True,  # PiGS starts DNS server thread after Schedule is built (PiGS ready for requests)
#      "DnsHosts"     : dict of "url" : "IP" host pairs for PiGS DNS server
#      "Server_IP"    : '0.0.0.0'
#          PiGS defaults to  0.0.0.0 listening on all active Interfaces eg eth0 wlan0 localhost 127.0.0.1
#          Use Server_IP to tell PiGS to listen on specific IP for port 80(web UI/epg requests) & 53(dns).
#          Sometimes needed in multiple NIC (wired/wifi) situations 
#
#    UI Updates
#    Added some link urls to SD & pigslog.txt, Added DNS enable checkbox and code behind it.
#    Added dropdown list for IconType / removed []Use CM Icons checkbox
#    Displayed listen IP near Version, updated the Help at bottom of page
#    Added note to remind user to do a dvr MGR on lineup or Icon changes
#    Added link to ./pigstatus on Config page
#
# V1.02rc V1.02a 2020-Sep-28
#    A beta version of 1.10 sent to testers.  Decided to bump to 1.10 since DNS is a big change
#
# V1.01 2020-Sep-22
#    Change DaysLimit to 0 to fetch the maximum amount of days from SD (was 16)
#    Add request counter to GetLinearSchedule call, used Cnt ==90 to print Done msg
#    Add PiGS Server IP and Port to pigs.json to allow IP to be customized w/o editing the Python code
#     "Server_IP"    : "0.0.0.0",
#     "Server_Port"  : 8000,
#    Convert canadian zips to uppercase in sd.py and fix bug in ls.py where I accessed empty stationList []
#    At start PiGS verifies PiGS lineups match SD account (eg if someone used epg123 or other prog to mess up SD)
#    Added error checking for SD server down status code 3000 that occurred 9-21-2020
#
# V1.00 2020-Aug-14
#    Builder can now receive params, signal the caller, and return results
#    Moved AddRemoveLineups, Preview, and SDLogin into Builder
#    Builder is really the SD-Resource-Manager, but Builder is its name
#    Disabled UI Save and 'Verify Login' buttons while PiGS processes the Commands
#    Next Rebuild Time (23hrs) is in pigs.json "RebuildSecs" : 23*60*60,
#    set permissions to 777 for myicons,cmicons,sdicons when they are created 
#
# status command updated with Builder info: 192.168.1.200/pigstatus
#
#---------------------------------------------------------------------------------------
# ToDo list
#
# scan code for ToDo's and fix or remove
#
# Could this code base be used with another guide provider?
# Probably.  The key tasks would be to :
#     1) Build the list of channel serices BuildServicesListNEW() using their data
#     2) Build the Linear schedule list BuildLinearSchedule() using their data
#     3) Plus a lot of UI stuff, like out how to login, configure the lineups(UI changes)
# The sequencing in this app was built based on the api provided by Schedules Direct
# If you were able to emulate the functions in SD.py (ie reformat everything there)
# using another providers data to provide equivalent results, it might go pretty 'smoothly'.      
#---------------------------------------------------------------------------------------

PIGS_LOG_FILE     = "pigslog.txt"
MAX_LOGFILE_BYTES = 150000  # 150K then rotate log file to next
LOGFILE_BACKUPS   = 3       # Current file and 3 backups

#---------------------------------------------------------------------------------------
# Json format configuration file

# PiGS can act as a DNS server.  This dict definition acts like a hosts file
# Pigs dns will translate name found here, but pigs dns used gethostbyname()
# so entries in /etc/hosts will be checked if an entry is not found here
#   HostDict is checked, if not found, then
#   /etc/hosts is checked, if not found, then
#   use the configured primary dns server (probaby router gateway)

HostsDict = {                               # PiGS DNS will replace '*' with the discovered IP or Server_IP 
    'epg.channelmastertv.com' : '*',        # fw 1.35 guide requests go here
    'roproxy-guide.echostarcdn.com' : '*',  # fw 1.34 and before guide requests go here
    'xpool.ntp.org' : '162.159.200.123',    # NTP update server requests, leading x causes dns forward
                                            # DVR+ uses these next 4 to determine if it is connected to the Internet
    'xechostar.com' :  '69.164.198.192',    # with the leading x dns is forwarding these IP's to a real DNS server.
    'xmicrosoft.com': '173.255.215.209',    # These are just some ntp server IPs addresses
    'xamazon.com'   : '104.236.116.147',    # DVR+ doesnt like it if I return IP's on the same network
    'xgoogle.com'   : '138.236.128.112',    # DVR+ doesnt access these addresses, it just wants DNS to succeed.
                                            # These to block firmware updates 
    'tr50.dishaccess.tv': '127.0.0.1',      # A tr50.dishaccess.tv CNAME stbfw.echodata.tv A 67.148.153.225
    'stbfw.echodata.tv' : '127.0.0.1'       # firmware update
}

CONFIG_FILE = 'pigs.json'

# Config json file is read into this global config dictionary at boot and on UI Save

dPiGS = None    # global config dictionary, empty until pigs.json is loaded
  
# This is the layout of the pigs.json master config file
# IF NEW KEYS ARE ADDED TO THE DICTIONARY, THEN ADD AN INIT LINE IN THE STARTUP SECTION
# SEARCH FOR "Auto-Upgrade" TO FIND THE LOCATION 

PigsConfigDict = {
                          # Required keys
  "sdusername"   : "",     # your SD username
  "sdsha1pwd"    : "",     # your SD password, sha1 encoded
  "Server_IP"    : "0.0.0.0",   # 0.0.0.0 means pigs will monitor port 80 on all IP's (wire & wired interfaces)
  "Server_Port"  : 8000,        # 8000 by default, use cmd line for port 80
  "RebuildSecs"  : 82800, # 23 hours = 23h*60*60 Next Build Time in seconds
  #"UseCMImages"  : True,  # use the pretty Channel Master / Rovi Icons.  Key removed in 1.02 for IconType
  "IconType"     : 1,     # 0=No Icons, 1=Sched Direct, 2=CM/Rovi Icons, 3=CutePigs???
  "CacheCMIcons" : True,  # keeps a local copy of the CM / Rovi Icons
  "CacheSDIcons" : True,  # keeps a local copy of the resized SD Icons
  "SaveFsSDIcons": False, # True Saves a copy of the FullSize SD icons in ./sdicons for external manipulation
  "SDImSize"     : [72,54], # 360w x 270 / 5 = 72 x 54 (json cant do Tuple, so will use a list)
  "SDImResample" : 3,     # Default to type 3.   Pillow Lib .resize() Resample type 0..5
  "SDImReduceGap": None,  # Pillow 7.0 Reducing Gap 
  "SdIconTheme"  : "dark", # choices are dark, gray, light, white
  "DefaultIcon"  : "pinkpig.png",  # if you guide icon is found... another could be "92x36box.png"   
  "PrependEpisode" : True,  # Prepend Episode Title, SeasonEpisode #'s, FirstAirDate
  "AppendCast"   : True,  # Append cast info to description
  "LoggingLevel" : 20,    # default logging level. DEBUG 10,INFO 20,WARNING 30, ERROR 40,CRITICAL 50
  "LoggingFile"  : PIGS_LOG_FILE,  # filename of log file, like pigs.logs.
  "LastSchdBuild": "?",   # Time of last schedule build, UTC
  "LastSvcBuild" : "?",   # Last time the SERVICES data was built (typ at boot, or after UI SaveSettings)
  "PiGS_Version" : PIGS_VERSION, # Version string so UI index.htm can display
  "PiGS_IP"      : '0.0.0.0', # Read-Only. IP Addr string so UI index.htm can display
  "PinkPigPos"   : 0,     # display 1 pink pig icon at pos N on the guide or on Chan x.y, use 0 to disable
  "AllowAnyZip"  : True,  # allow any text in Zip box.  False = USA 5 digit, numbers only
  "PigsDoesDNS"  : True,  # PiGS acts as a DNS Server for DVR+.  Forward requests for msft,google,amzn, etc
  "PigsDelayDNS" : True,  # PiGS starts DNS server thread after Schedule is built  
                          # Development keys
  "MGRnewSched"  : True,  # Manual Guide Refresh forces Rebuild of SD Schedule
  "debugToken"   : "",    # if this key exists, it over-rides the SD token request
  "DaysLimit"    : 0,     # Num days fetched from SD (DVR+ can only use 14), 0 for Max Days, pad to 16 to prevent 403 errors
                          # DaysLimit was used to limit day to 2 or 3 for fast builds during development.
  "ChanFilter"   : "0.0", # not used - used to filter channel to one on DVR+ schedule request
  "LogHttpReq"   : False, # debug enable annoying http logs on each send
  "SaveJson"     : False, # save the PiGS Channel Lineup sent to DVR+ in servicedetailsmsg.json for troubleshooting
  "TimeZones"    : [      # This gets copied into the DVR+ message response headers
      {
        "StartDateTime" : "2020-03-08T08:00:00Z",
        "EndDateTime"   : "2020-11-01T07:00:00Z",
        "Offset" : "???"  # ??? will be initialized with users local as found in their system
      }
    ],
  "DnsHosts" : HostsDict,
                          # Finally, list of users lineups, limited to 4 by Schedules Direct
  "MyLineups"    : []     # List of LineupDef objs with users favorite stations flagged
                          # Multiple Lineups allow user to pull schedule data from far away stations.
                          # e.g. their local zipcode might not include all the stations they can receive.
                          # MyLineups will be used to build the ServiceDetails station List DVR+ requests  
}

#---------------------------------------------------------------------------------------
# UI Settings of config
# A full copy of the PigsConfigDict is sent to the UI index.htm for manipulation
# Some is for display only, but some is for getting the users settings.
#
# The following PigsPreserveSettings dict is a list of keys that will be preserved
# when the user hits Save Settings.
#
# Anything in this list becomes essentially READ-ONLY - UI cant change it.
# E.g, if a guide update occurred while the user had the UI open and the
# LastSchdBuild time had been updated, we want to preserve that key value since
# it occurred after the UI was opened.
# Another example is the sha1 password.  Initally it does not exist in the config file.
# User presses verify login - UN & PW is sent to the server and the  sha1 is created
# and saved in the server conig, but not sent back to the UI so the UI data is 'out of date,'
# so the server side sha1 is 'Preserved' when the user clicks save settings.
# The key names here must match those in PigsConfigDict exactly
# The values in PigsPreserveSettings do nothing - Just place holders.
# I suppose I could just make a straight List [] object of key names instead of a dict object.

PigsPreserveSettings = {
  "sdsha1pwd"    : "",  # your SD password, sha1 encoded
  "LastSchdBuild": "",  # Time of last schedule build
  "LastSvcBuild" : "",  # Last time DVR+ did a ZIP,SERVICES request
  "PiGS_Version" : "",  # Version string for loggin
  "PiGS_IP"      : "",  # IP string for loggin
  "debugToken"   : "",  # this key over-rides the token request
  "DaysLimit"    : "",  # limit number of schedule days fetched from SD
  "ChanFilter"   : "",  # not used - used to filter channel to one on DVR+ schedule request
  "LogHttpReq"   : "",  # debug enable annoying http logs on each send
  "SaveJson"     : ""   # save DVR+ json request body
}

# A lineupDef is created when the User clicks 'Preview Lineup' with a valid ZIP
# 
# It will be created by PiGS and sent to the UI but exist only on the UI side until the user
# clicks 'Add This Channel Lineup' (added to MyLineups[]) followed by Save Settings (send to server)
# 'Add..." action will add the 'Preview" to the UI copy of config MyLineups[].
# So the MyLineups array can/will change size from what was sent to the UI
# 
# To commit UI changes to the Server, they must click 'Save Settings' to send to PiGS
#
# SD allows for up to 4 lineups
# LineupDef dict is one user lineup: the LineupName for UI display, and list of favorite chans.
# MyLineups[] is a list of LineupDef's in pigs.json

LineupDef = {
  "LineupName" : "LU-Name",   # SD Name for the lineup
  "ChannelList": []           # List of ChannelDef objs. It will include every channel in the lineup
  }

# The ChannelDef object below is for reference only.  The first 4 keys come from SD preview call.
# The SD preview will send the channel,name,callsign,and affiliate keys, I will add the favorite key
# after getting the results from the SD command.

# When the user saves the lineup, the DVR+ services list will be created using the
# channels marked as favorites and info obtained from SD station Map command

# These are the key names sent from the SD lineup preview results, with the addition of
# my favorites key.  we will loop through and add the favorite key after getting the data
# from SD so the favorite key exists for the UI and has an initial value

ChannelDef = {
      "channel"  : "3.1",    # with example values.
      "name"     : "WISCDT (WISC-DT)",
      "callsign" : "WISCDT",
      "affiliate": "CBS",
      "favorite" :  True      # by default.
}

#---------------------------------------------------------------------------------------
# References for HttpServer object
#https://github.com/python/cpython/blob/3.8/Lib/http/server.py

from http.server import BaseHTTPRequestHandler, SimpleHTTPRequestHandler

from http.server import HTTPServer
#from http.server import ThreadingHTTPServer

from io import BytesIO
from urllib.parse import parse_qs, urlparse

# CM/DVR+ bases channel lineups on Service ID's  Multiple ZIP's might map to one service ID
# PiGS will be configured up front using Schedule Direct lineups 
# Going to use a 'dummy' service ID of 99999 everywhere.
# Every DVR+ zip request will get a service id of 99999
# The 99999 number is important b/c it must be consistent across all DVR+ service calls,
# listing calls, and program airings or DVR+ wont display it.
# The actual channel lineups will be configured in the UI / via Schedules Direct

PiGS_SERVICEID = 99999

#running starting a program at bootup on a Pi
#https://www.dexterindustries.com/howto/run-a-program-on-your-raspberry-pi-at-startup/

#SERVER_HOST = '0.0.0.0'    #'localhost' did not work for outside requests
#SERVER_PORT = 8000         # to run with port 80 cmd line: "sudo pyton3 pigs.py 80"

#https://docs.python.org/3.7/library/logging.html?highlight=logging#module-logging
#https://www.blog.pythonlibrary.org/2014/02/11/python-how-to-create-rotating-logs/

import logging
import logging.handlers
import time
from   datetime import datetime, timedelta
import json
import gzip  # jsgz = gzip.compress(js)
import sched
import sys
from   copy import copy
import copy

# Functions to send data to DVR+
from sv import GetServiceIdFromZip, GetServiceDetails, BuildServicesListNEW, GetServiceList, GetSVCBuildInfo
from ls import BuildLinearSchedule, GetLinearSchedule, GetDVRStates, GetLSBuildInfo
from sd import GetSha1Password, SDGetToken, SDGetLineupPreview, SDAddDeleteLineups

from pigsdns import PigsDnsStart, PigsDnsStop, PigsDnsAlive, PigsDnsStatus

# utility functions
from ut import LocalTimeStr
from ut import VersionStringToFloat

bLogHttp = False  # a flag to stop some annoyingly large httplib debug messages

#https://docs.python.org/2/library/os.path.html
import shutil
from os import path
import os

#=========================================================================
# Save config file, keep a history of 5 backups
# https://docs.python.org/3/library/shutil.html
# https://www.w3schools.com/python/python_file_remove.asp
# https://docs.python.org/3/library/shutil.html
# https://docs.python.org/3/library/os.path.html
# https://docs.python.org/3/library/os.html#module-os
# I wonder if there is some type of Rotation Save function in Python

CFG_HISTORY = 5

def SaveSettingsFile(configDict, configFile) :
    
    # separate file path/name from its extention
    root, ext = path.splitext(configFile)

    # delete the oldest log file #5
    dst = root+str(CFG_HISTORY)+ext
    if os.path.exists(dst):
        os.remove(dst)
    
    # then rename files to keep history
    # ren 4 to 5, 3 to 4, 2 to 3, etc
    for i in range(CFG_HISTORY-1,-1,-1) : # start, stop 1 before, step
        src = root+str(i)+ext if (i>0) else root+ext  # strip '0' for current file 
        dst = root+str(i+1)+ext # to 5
        if os.path.exists(src):
            os.rename(src, dst)  
        #src = path.realpath("guru99.txt");
        #os.replace(src, dst, *, src_dir_fd=None, dst_dir_fd=None)

    # and write config file to storage 
    # pretty json with indent=2.  sorted is ugly and ungrouped
    with open(configFile, "w") as outfile:    
        json.dump(configDict, outfile, sort_keys=False, indent=2) 
    
    outfile.close()

    #TodDo  # seems to be only 644 so user cant delete file. do i want that???


#---------------------------------------------------------------------------------------
# The schedule build can take a long time, up to a minutue or two (or even more).
# To allow the web server to respond to multiple DVR requests, UI commands, and
# web pigstatus requests in real-time, the build will be run in a separate thread.  
# If a request comes in while the Build is in progress, the webserver will report
# a http 503 server busy error.
# This will allow the webserver to respond to all requests instead of ignoring requests
# when the build is in progess.
# It appears requests were stacking up and confusing the http.server resulting in
# low level "broken pipe" errors.
#
# A Build Timer thread is is used to build the schedule about 1 hour before a typically requests it.


import queue
import threading

BuilderThread   = None               # Thread Handle
BuilderQueue    = queue.Queue()      # Queue to send Buildxxx commands to Builder
BuildAbortEvent = threading.Event()  # Event used to abort Build if needed.  isSet()  set()  clear()  wait(timeout)
SDResourceLock  = threading.Lock()   # Schedules Direct Resource Lock, only 1 command can use it at a time

BuildLogin     = "Login"
BuildDelay     = "Delay"
BuildServices  = "Services"    
BuildSchedule  = "Schedule"
BuildTerminate = "Terminate"
BuildAddRemove = "AddRemove"
BuildPreview   = "Preview"

# Commands to the Builder are sent via a dict structure. 'Command' key is required.
# The remaining keys are defined by the command
BldCmd = {
    'Command' : None,    # one of the Above Buildxxxxx commands
    'Params'  : None,    # Params to pass in.  See BuildAddRemove for example
    'Event'   : None,    # If the caller needs to be signaled when result is ready. See BuildPreview for example
    'Result'  : None     # Results to send back to caller. A dict works good.  See BuildPreview for example
  }


# Handles to 2 build timers - One started by DVR, and one as a Watchdog if DVR never requests data
WDogBuildTimer  = threading.Timer(1,None)   # watchdog build timer 
DvrBuildTimer   = threading.Timer(1,None)
NextBuildTimeStr= ""
AutoBuildTimeStr= ""

# This is a watchdog timer - it ensures that the schedule gets built at lease every 24 hrs
# This timer is activated right after the DVR build timer fires in case a DVR never contacts pigs to reset timer

def DoWDogBuildTimer():
    logging.warning("THRD:DoWDogBuildTimer: has fired. Where are the DVR's?")
    SetDvrBuildTimer(1)       # Trigger the normal Timer build process 1 second from now
    # Kicking off the Build timer will also reset this Auto build timer.

# DoTimerBuild is called when the DvrBuildTimer thread fires
# The Timer thread is dead after the Build command is Queued
# The build timer will be reset when a dvr makes a schedule request.

def DoTimerBuild():
    global WDogBuildTimer
    global AutoBuildTimeStr

    logging.info("THRD:DoTimerBuild: has fired. Waiting for SDResourceLock Lock...")

    # Do a hard wait for the SD Lock here.  This waits until the Build Queue is empty
    if SDResourceLock.acquire(True) :  
        logging.info("THRD:DoTimerBuild: Acquired SDResourceLock Lock.")
        
        # added in V1.02.
        # If ServiceList did not get built due to some error,  we will build it here.
        # If after the BuildSchedule, services is not build, the WatchDogtime will be set for 30-minutes
        if not GetServiceList() :
            logging.warning("THRD:DoTimerBuild: No Service Lineup, Queuing BuildServices")
            BuilderQueue.put( {'Command':BuildServices} )
        
        logging.info("THRD:DoTimerBuild: Queuing BuildSchedule.")
        BuilderQueue.put( {'Command':BuildSchedule} )
    
    #SetWDogBuildTimer(24*60*60)   # 1.02, this is now done after a scheudle build completes
    
    return 

def SetWDogBuildTimer(delaySeconds) :

    global WDogBuildTimer, AutoBuildTimeStr
    
    # When DoTimerBuild() ends, the DvrBuildTimer thread is dead.
    # We expect a dvr to come along within the next hour, to reset this timer, but if it does not...
    # If the dvrs are offline, eg the network is down, the sched could get really stale.
    # The Build timer was set for 23 hrs after the a dvr made a schedule request, we now expect
    # that dvr to make a request within an hour.  We will set a 24 hour timer 'watchdog' timer
    # so that if dvr doesnt ask in the expected time, we will do a build anyway.
    # The WDogTimer will get reset right here each time it fires.

    WDogBuildTimer = threading.Timer(delaySeconds, DoWDogBuildTimer)
    WDogBuildTimer.setName('Auto Build Timer')
    WDogBuildTimer.setDaemon(True)         # thread is automatically killed when program exits
    WDogBuildTimer.start()
    
    at  = datetime.utcnow()                # return the local clock time when the timer will fire
    at += timedelta(seconds=delaySeconds)
    AutoBuildTimeStr,tz = LocalTimeStr(at, showSeconds=True)

    logging.info("THRD:SetWDogBuildTimer: Auto Build time set for %s.", AutoBuildTimeStr)

    return AutoBuildTimeStr


def SetDvrBuildTimer(delaysecs) :
    global DvrBuildTimer
    global NextBuildTimeStr
    global WDogBuildTimer
    
    # If the auto build timer is set, cancel it since we are scheduling a build
    try :
        if WDogBuildTimer.is_alive() :
            WDogBuildTimer.cancel()
    except :
        logging.warning("THRD:SetDvrBuildTimer: Exception canceling WDogBuildTimer.")

    try :  # if DvrBuildTimer is currently active, kill it and restart with a new time
        if DvrBuildTimer.is_alive() :
            DvrBuildTimer.cancel()
    except :
        logging.warning("THRD:SetDvrBuildTimer: Exception canceling DvrBuildTimer.")
    
    # have to create a new Timer thread each time, cant reuse or reset thread timers
    DvrBuildTimer = threading.Timer(delaysecs, DoTimerBuild)
    DvrBuildTimer.setName('DVR Build Timer')
    DvrBuildTimer.setDaemon(True)             # thread is automatically killed when program exits
    DvrBuildTimer.start()
    
    bt  = datetime.utcnow()                # return the local clock time when the timer will fire
    bt += timedelta(seconds=delaysecs)
    NextBuildTimeStr,tz = LocalTimeStr(bt, showSeconds=True)
    
    return NextBuildTimeStr

# Builder Thread Functions will generally use global variables, but some pass in parameters and return results

def Builder() :
    
    global BuilderQueue
    global BuildAbortEvent
    global SDResourceLock

    Terminate = False
    
    while not Terminate :  # Builder continues until terminated (nothing terminates it right now)
        
        # Check if someone is requesting a Builder abort (UI save settings an do this)
        if BuildAbortEvent.is_set() :
            logging.info("THRD:BUILDER: Abort Request. Flushing Command Queue.")
            BuildAbortEvent.clear()
            while not BuilderQueue.empty() :
                BuilderQueue.get()
                BuilderQueue.task_done()
        
        # Builder keeps the SDResourceLock until all commands have been processed
        # If Queue Len is 0 andBuilder has the lock, then release the SD Resource Lock
        if BuilderQueue.empty() and SDResourceLock.locked() :
            logging.info("THRD:BUILDER: Queue is Empty, releasing SDResourceLock\n")
            SDResourceLock.release()
            
        # Before placing a command in the Queue, the initiator requested the SDResourceLock
        # The caller could have submitted multiple build commands
        # Get a build command from Queue.  this Thread will wait/block here for a command
        
        dBuildCmd = BuilderQueue.get()  # its actually a Build Dictionary with 1 or more keys
        buildCmd = dBuildCmd['Command']
        logging.info("THRD:BUILDER: Command is %s",buildCmd)

        # First handle non-SD commands that dont need a Token

        if buildCmd == BuildDelay :  # this one just helps it appear things happen in order.  cosmetic only
            time.sleep(1)

        elif buildCmd == BuildTerminate :
            logging.info("THRD:BUILDER: Builder Thread Terminating...")
            Terminate = True

        else : # its an SD command that needs a token
            
            un = dPiGS['sdusername']     # default: use username and pwd from pigs.json
            pw = dPiGS['sdsha1pwd']
        
            if buildCmd == BuildLogin :  # special case for initial login, un/pw not in dPiGS yet
                un = dBuildCmd['Params']['un']
                pw = dBuildCmd['Params']['pw']

            # attempt a GetToken with username and sha1pwd to talk to the SD API
            code, tokencode, errj = SDGetToken( un, pw, dPiGS.get('debugToken'))

            if buildCmd == BuildLogin :  # special case for initial login, return token status
                
                # send the result back as a small dict object
                dBuildCmd['Result'] = {'code':code+tokencode, 'errj':errj}
                dBuildCmd['Event'].set()  #signal waiting process results are ready
            
            if (code == 200 and tokencode == 0) :  # Token get successful -
                
                if buildCmd == BuildLogin :  # nothing to process
                    pass
                
                elif buildCmd == BuildAddRemove :
                    try :
                        lineups = dBuildCmd['Params']
                        SDAddDeleteLineups(lineups)
                    except :
                        logging.warning("THRD:BUILDER: BuildAddRemove Lineup Exception occurred!")
                    
                elif buildCmd == BuildPreview :
                    try :
                        code, lujson, luname = SDGetLineupPreview( dBuildCmd['Params'] )  #
                        # send the result back as a small dict object
                        dBuildCmd['Result'] = {'code':code, 'lujson':lujson,'luname':luname}
                        logging.info("THRD:BUILDER:Preview: Setting callers Event flag.")
                        dBuildCmd['Event'].set()  #signal results are ready
                    except :
                        logging.warning("THRD:BUILDER: Build Preview Exception occurred!")

                elif buildCmd == BuildServices :
                    try :
                    #if True:
                        BuildServicesListNEW(dPiGS, PiGS_SERVICEID)

                        # for debug, save the json (large) message sent to dvr+
                        if dPiGS.get('SaveJson') :
                            logging.info('THRD:BUILDER: Saving ServiceDetailsMsg to '"'servicedetailsmsg.json'"' ')
                            sdjson = GetServiceDetails(dPiGS, PiGS_SERVICEID, dPiGS['TimeZones'])
                            file = open('servicedetailsmsg.json', "w")
                            file.write(sdjson)
                            file.close()
                    except :
                    #else:
                        logging.warning("THRD:BUILDER: Build Service Lineup Exception occurred!")
        
                elif buildCmd == BuildSchedule :
                    try :
                        BuildLinearSchedule(dPiGS, PiGS_SERVICEID, BuildAbortEvent)
                    except :
                        logging.warning("THRD:BUILDER: Build Schedule Exception occurred!")
                        
                    # If PiGS was not able to build the service list at boot, we catch it here
                    # and schedule a timer build 30 minutes from now
                    if not GetServiceList() :
                        SetWDogBuildTimer(30*60)     # 30 mins from now
                    else :                           # 
                        SetWDogBuildTimer(24*60*60)  # if buld success schedule 24 hours from now
                        
                        # Start DNS after schedule is built so DVR's dont ask too early
                        if dPiGS['PigsDoesDNS'] and not PigsDnsAlive() :  # and dPiGS['PigsDelayDNS'] assume delayed
                            logging.info("THRD:BUILDER: Starting DNS...")
                            PigsDnsStart(dPiGS['Server_IP'], IPaddr, dPiGS['DnsHosts'])                    

            else :
                # this might happen if user has not set their SD UN/PW
                logging.warning("THRD:BUILDER:%s GetToken Error(%d) %s", buildCmd, code+tokencode, errj)
        
        BuilderQueue.task_done()
        
        # take a 200mS nap. Other code sometimes submits multiple SD commands.  Want to make sure
        # the main thread had time to submit all the commands it wanted to
        
        time.sleep(0.200)
        
    logging.info("THRD:BUILDER: Builder Thread Terminated!")
    
    return


#---------------------------------------------------------------------------------------
# References
# https://blog.anvileight.com/posts/simple-python-http-server/
# This reference also discusses SSL support, althought DVR+ does not do SSL
# Subclass because I need to serve some files (index.htm, log file, images) but also
# need to handle the GET and POST requests BEFORE passing on the Simple handler

#Server_error_responses https://developer.mozilla.org/en-US/docs/Web/HTTP/Status
ServerErrorCode = 503  #

class PiGS_HTTPRequestHandler( SimpleHTTPRequestHandler ):
    
    global bLogHttp
    global dPiGS
    global PigsStartTime

    # Scheudle Builder communication
    global BuilderQueue
    global BuildAbortEvent
    global SDResourceLock
    global BuilderThread
    
    # Scheudle Builder Commands
    global BuildDelay    
    global BuildServices  
    global BuildSchedule 
    global BuildTerminate
 
 
    #https://realpython.com/intro-to-python-threading/
    #def __init__(self) :
    #SimpleHTTPRequestHandler.__init__(self)
    #super(PiGS_HTTPRequestHandler, self).do_GET()
    # super().__init__()
    #    self.__init__()
    #    self.BuildThread = threading.thread(BuildLinearSchedule, (dPiGS, PiGS_SERVICEID), deamon=True)
    #    self.BuildThread.start()
    #Thread.__init__()
    #    self.sdlock = threading.Lock()
        
    # self.client_address .server .command .path .request_version .headers)
    
    def SendCommonHeaders(self):

        # HTTP/1.1 200 OK\r\n
        # Headers sent by CM servers to DVR+, We will emulate
        #SimpleHTTP sends these two automatically
        #self.send_header("Date:", "Sat, 02 May 2020 16:04:00 GMT") #Sat, 02 May 2020 16:04:00 GMT\r\n
        #self.send_header("Server:", "nginx/1.10.3 (Ubuntu)")
 
        # These emualate what was wiresharked from DVR+.  They could be changed
        self.send_header("X-AspNet-Version", "4.0.30319")
        self.send_header("X-Mashery-Responder", "prod-j-worker-us-west-1b-22.mashery.com")
#        self.send_header("Cache-Control", "private, max-age=0")
        self.send_header("Cache-Control", "no-store")
#        self.send_header("Connection", "keep-alive")
        self.send_header("Vary", "Accept-Encoding")
        self.send_header("Content-Type", "application/json")
        
        #These will be set by individual commands
        #self.send_header("Content-Length:", "359")  # size of payload, gzipped data
        #self.send_header("Content-Encoding:", "gzip")
        
    # this little function silences the logging messages sent from
    #      self.send_response()  log_message(format, ...) log_error(format, ...)
    # https://docs.python.org/2/library/basehttpserver.html
    # https://stackoverflow.com/questions/3389305/how-to-silent-quiet-httpserver-and-basichttprequesthandlers-stderr-output

    def log_message(self, format, *args):
        if bLogHttp :                    # a key allows me to block or show these messages
            logging.info(format, *args)  # comment this out to completely silence http:lib
            #logging.warning(format, *args)
            pass
        return
    
    def do_GET(self):   # I get to handle GET's before passing up to parent 'fileserver'
        
        global dPiGS    # uses the global dictionary
        global DvrBuildTimer
        global NextBuildTimeStr
        global AutoBuildTimeStr
                                # https://docs.python.org/3.7/library/urllib.parse.html
        o=urlparse(self.path)   # break url into scheme,netloc,path,params, query,fraqment
        qs=parse_qs(o.query)    # parse the query parameters into a dictionary
        
        #logging.debug("GET: Path: %s\nHeaders:%s", str(self.path), str(self.headers))
        #logging.debug("GET: Path: %s", str(self.path))#, str(self.headers))

        # Handle DVR+ of Get service code from Postal Code
        # The DVR+ zip is going to be ignored - we will return 99999 as the Service Code
        
        # If we wanted to send commands to the server from DVR+, we could 
        # use the zip codes as 'commands'.  eg 5000 could be force SD refresh.
        # For example, set the zip to 50000, then do a guide refresh.
        # The 50000 would be interpretted here and start the download
        # Could require one of these to set the daily download time.
        # after the command, the zip should be set to 99999 on dvr+
        # if user did not reset zip to 99999 the cmd would execute each time dvr did a zip request.
        # We'd need to have a Latch flag (did it), and a reset (another zipcode) for this to work
        
        # Get time of request, convert to local time - easier to match up with what we see on dvr stbhealth page
        lts, tz = LocalTimeStr(datetime.utcnow(), showSeconds=True)
        
        if ("postalcode" in o.path): # DVR+ asking for GET of service ID
            
            logging.info("MAIN:**** Postalcode request from %s at %s.", self.client_address, lts )
            
            # Postalcode only happens on a Manual Guide refresh, or when user does some things on the DVR
            # We will cancel the scheduled build time so that the build time syncs with the next dvr???
            
            self.send_response(200)
            self.SendCommonHeaders()
 
            # We could parse out the postalcode from the URL sent by DRV+ but we are not using it
            # when DVR+ sends a zipcode, we always return a service code of 99999
            # Send the current TimeZone info for incorporation to DVR+
            # Get a json response object to send to DVR+
            
            srjson = GetServiceIdFromZip("", PiGS_SERVICEID, dPiGS['TimeZones'])
            
            # This is a small message... no encoding needed
            #self.send_header("Content-Encoding:", "gzip")

             # size of payloadr
            self.send_header("Content-Length", str(len(srjson))) 
            self.end_headers()                      
            self.wfile.write( srjson.encode() )
           
        elif ("servicedetails" in o.path): #DVR+ asking for GET of service ID
            
            # this message can't happen without DVR+ first calling postalcode, so we dont
            # have to check for the build state.
            # Todo, well maybe it could from another DVR????

            logging.info("MAIN:**** Service Details request from %s at %s.", self.client_address, lts )
            
            # if network was not ready at boot, the service list may not have been built
            if not GetServiceList() :
                if SDResourceLock.acquire(False) :  # if we got the lock, start the build                
                    logging.warning("MAIN:SERVICEDETAILS: No service list, Initiating scheudule build process...")
                    SetDvrBuildTimer(1.0)
                    #logging.warning("MAIN:SERVICEDETAILS: No service list, building service list & schedule...")
                    #BuilderQueue.put( {'Command':BuildServices} )
                    #BuilderQueue.put( {'Command':BuildSchedule} )
                else :
                    # seems like this could never happen - cant be building anything if there is no service list
                    logging.warning("MAIN:SERVICEDETAILS: PiGS is busy rebuilding schedule! (%S)", ServerErrorCode)
                    
                self.send_error(ServerErrorCode, message="PiGS is busy building the Schedule")
                return

            # received headers Accept: */*  Accept-Encoding: gzip,deflate
            self.send_response(200)
            self.SendCommonHeaders()
            
            # Send the Channel lineup list of users Favorite Stations/Channels
            # these have been preconfigured in the UI and are in pigs.json       

            # Get service details and Send to DVR+
            # The ServiceDetails list was built at boot and when user saved settings
            
            # Send DVR+ a service list that was built from SD information
            # This contain the channel numbers, image links, station ID's, etc
            # DVR+ keeps a local copy of this
            
            sdjson = GetServiceDetails(dPiGS, PiGS_SERVICEID, dPiGS['TimeZones'])
            
            # gzip encocded response
            sdgz = gzip.compress( sdjson.encode() )  #to utf-8
            #logging.debug("sdlen(%d)  gzlen(%d)\n", len(sdjson), len(sdgz))

            #gzip encoded response
            self.send_header("Content-Encoding", "gzip")
            self.send_header("Content-Length", str(len(sdgz))) 
            self.end_headers()
            self.wfile.write( sdgz )

        elif ("linearschedule" in o.path): # DVR+ is asking for a GET of listings

            logging.info("MAIN:LINSCHED **** Program Sched Req from %s at %s", self.client_address, lts )
            
            # if network was not ready at boot, this list may not have been built
            if not GetServiceList() :
                
                if SDResourceLock.acquire(False) :  # if we got the lock, start the build                
                    logging.warning("MAIN:LINSCHED: No service list, Initiating scheudule build process...")
                    SetDvrBuildTimer(1.0)
                    #BuilderQueue.put( {'Command':BuildServices} )
                    #BuilderQueue.put( {'Command':BuildSchedule} )
                else :
                    logging.warning("MAIN:LINSCHED: PiGS is busy rebuilding schedule! http(%s)", ServerErrorCode)

                self.send_error(ServerErrorCode, message="PiGS is busy building the Schedule", )
                return
            
            # could happen if another dvr made a request just after the Build Timer fired
            if SDResourceLock.locked() :  # test the SD lockif                
                logging.warning("MAIN:LINSCHED: PiGS is busy rebuilding schedule! http(%s)", ServerErrorCode)
                self.send_error(ServerErrorCode, message="PiGS is busy building the Schedule" )
                return

            # In normal daily updates DVR+ just jumps right in with a schedule fetch since
            # it knows its service details already

            # This is how the schedule is built on the fly...
            # Schedule the time of the next schedule build 23 hrs from now, just before this dvr needs it.
            # The other dvrs will just use the existing data b/c its already been built
            # Use the first request from a dvr to schedule the next build time
            
            if qs["inprogress"][0]=='true' :   # if this is a DVR's first request of the update
                if not DvrBuildTimer.is_alive() :  # and the Build timer is not set
                    delaytime = dPiGS["RebuildSecs"]  # then set it for 23 hrs in seconds from now
                    bts = SetDvrBuildTimer(delaytime)
                    logging.info("MAIN:LINSCHED: Next Build time is scheduled for %s", bts)
                    # Once the Timer fires and the Build command is Queued, the DvrBuildTimer thread will be dead

            self.send_response(200)
            self.SendCommonHeaders()
            
            #debugging to just update 1 channel for speed on DVR+
            chf = dPiGS.get('ChanFilter')
            
            # Send Linear Schedule to DVR+  qs = has the request params sent by DVR+
            lsjson = GetLinearSchedule(dPiGS, qs, PiGS_SERVICEID, \
                                       self.client_address[0], dPiGS['TimeZones'], chf )

            # This is a very large response - typically 4hours x many channels of data
            lsgz = gzip.compress( lsjson.encode() )  #to utf-8
            self.send_header("Content-Encoding", "gzip")
            logging.debug("LinearSched: lslen(%d)  gzlen(%d)\n", len(lsjson), len(lsgz) )
            
            #DVR+ server does chuncked http encoding. I dont know how to do this, or
            #even if we need to do chunked encoding - seems to work without it.
            #or does this httpserver does it automatically??  Does not appear so.
            #self.send_header("transfer-encoding", "chunked")

            self.send_header("Content-Length", str( len(lsgz) )) 
            self.end_headers()
            self.wfile.write( lsgz )
            
        elif ("pigstatus" in o.path): #GET of 

            logging.info("MAIN:PIGSTATUS: PiGS Status request from %s at %s.", self.client_address, lts )

            self.send_response(200)
            self.SendCommonHeaders()

            # add some more PiGS info to the results
            pigstatus = {}
            
            sts, lz = LocalTimeStr( PigsStartTime, True )
            pigstatus["PiGS Start Time"] = sts
            
            uptime = datetime.utcnow() - PigsStartTime
            pigstatus["   PiGS Up-Time"] = str(uptime)
            pigstatus["   PiGS_Version"] = dPiGS['PiGS_Version']
            pigstatus["  Builder Alive"] = BuilderThread.is_alive()
            dnsalive,rtime = PigsDnsStatus()
            lrt,lz = LocalTimeStr(rtime, showSeconds=True, showDate=False, showTimeZone=False)
            pigstatus["    DNS Enabled"] = str(dnsalive) + ', Last Req: ' + lrt # just the time
            pigstatus[" PiGS Directory"] = os.getcwd()
            
            svcStatusDict = GetSVCBuildInfo()
            pigstatus.update(svcStatusDict)  # and merge it into to pigstatus dict
            
            buildDict = GetLSBuildInfo();  #
            pigstatus.update(buildDict)  # and merge it into to pigstatus dict
            
            # Then add the Next Build Time
            msg1 = NextBuildTimeStr if DvrBuildTimer.is_alive() else  "Not Scheduled."
            pigstatus['Schedule Build Info']['NextBuildTime'] = msg1

            msg2 = AutoBuildTimeStr if WDogBuildTimer.is_alive() else "Not Scheduled."
            pigstatus['Schedule Build Info']['AutoBuildTime'] = msg2
                
            pigstatus["DVR Status List"] = '------------------------'
                           
            # Get the status of all DVR's PiGS has updated
            dvrStateD = GetDVRStates();  #
            pigstatus.update(dvrStateD)  # and merge it into to pigstatus dict
            
            dvrjson = json.dumps( pigstatus, indent=2 )
            
            #unencoded response 
            self.send_header("Content-Length", str( len(dvrjson)) ) 
            self.end_headers()
            self.wfile.write( dvrjson.encode() )
        
        elif ("UIsettings" in o.path): #GET of 

            logging.info("MAIN:GET UIsettings...")

            self.send_response(200)
            self.send_header("Content-type", "application/json")
            self.end_headers()

            # export dict to json string, then encode str to bytes object
            jstr = json.dumps( dPiGS )   # Send everything to the UI
            
            response = BytesIO()
            response.write( jstr.encode() )
            self.wfile.write( response.getvalue() )
            return
        
        elif ("lineupPreview" in self.path) :  # UI requesting a SD Lineup Preview based on a Zip

            # Preview needs the SD resource lock b/c it uses the SD token.
            # if a schedule build was in progress this UI call would kill the build

            if not SDResourceLock.acquire(False) :  # if we get the lock, we can do the preview
                logging.warning("MAIN:PREVIEW: Can't Preview.  PiGS is busy rebuilding schedule!")
                self.send_error(ServerErrorCode, message="PiGS is busy building the Schedule. Please Retry.")
                return

            # Build a LineupDef to send to the UI fo display
            # LineupDef = {
            #   "LineupName" : "USA-OTA-53593",   # SD Name for the lineup
            #   "ChannelList": []                 # List of ChannelDefs.
            # }                                   # It will include every channel in the lineup

            # Builder will signal us when Preview data is ready
            PreviewReadyEvent = threading.Event() 
            PreviewReadyEvent.clear()

            # create a build command 
            pcmd = copy.copy(BldCmd)
            pcmd['Command'] = BuildPreview
            pcmd['Params']  = qs['postalcode'][0]
            pcmd['Event']   = PreviewReadyEvent
            
            logging.info("MAIN:GET:lineupPreview: Queuing BuildPreview for %s.  Waiting for Result...", qs['postalcode'][0])
            BuilderQueue.put( pcmd )
            
            # this command could take a long time if the Preview has to be built by the server.
            # it might not even complete on teh SD server in time, the user just has to retry.
            # The PiGS webserver is busy during this wait and cant respond to other commands
            
            PreviewReady = PreviewReadyEvent.wait(60) # wait here for up to 60 seconds
            
            if PreviewReady :
                logging.info("MAIN:GET:lineupPreview: Result is ready.")
                # Get the results
                d      = pcmd['Result']   # get reference to the result dictionary 
                code   = d['code']
                lujson = d['lujson']
                luname = d['luname']
            else :
                logging.warning("MAIN:GET:lineupPreview: Timeout waiting for Preview Result.")
                code = 503
                lujson = None
                luname = "Timeout"
            
            # Call Lineup Preview command which makes 2 SD calls to get the data
            # Request the lineup preview, we can get a 200 with an error block with a code in it
            # There is also a differnt type of error response with a non-zero code.
            # we dont know which we will get
                       
            if code == 200 :
                # check for error code in json data, it might be a 2107 Queued not ready error
                # check for an error code, the preview response doesnt have an error code
                
                resultcode = 0
                
                # SD returns List[] for preview or dict {} for error / status response
                if type(lujson) is dict :
                    resultcode = lujson.get('code',0)
               
                if resultcode == 0 :
                    
                    # The returned lujson IS the LineupDef ChannelList, but the ChannelDef's
                    # dont include the favorite key
                    # Create the 'favorite' key and default to True for all channels
                    for c in lujson :
                        c['favorite']=True
                        
                    # create the complete response message to send to the UI
                    msg = copy.copy(LineupDef)    # Start with a copy of my custom response
                    msg['LineupName']  = luname;  # Add the SD lineup name (for display on the UI side)
                    msg['ChannelList'] = lujson;  # append the lineup list of ChannelDefs with Fav key
                    msgj = json.dumps(msg)        # dict to json
                else :
                    logging.info("MAIN:PREVIEW: return code %d", resultcode)                
                    msg = {'code':2107, 'message':"The lineup is being generated, please retry."}
                    msgj = json.dumps(msg)
            else :
                logging.info("MAIN:PREVIEW: Error %d %s %s", code, lujson, luname)                

                
            if lujson == None :
                msg = {'code':code, 'message': luname }
                msgj = json.dumps(msg)

            # forward the SD response code to the UI
            self.send_response(code)
            self.SendCommonHeaders()
            
            # gzip encocded response
            lugz = gzip.compress( msgj.encode() )  #to utf-8
            
            self.send_header("Content-Encoding", "gzip")
            self.send_header("Content-Length", str( len(lugz) )) 
            self.end_headers()

            logging.debug("  GET lineupPreview: lulen(%d)  gzlen(%d)\n", len(msgj), len(lugz) )

            self.wfile.write( lugz )
            return
             
        else: # we have not handled this GET.  Let SimpleHTTP handle file requests
            logging.info("MAIN:GET: Calling Parent for file request: %s", o.path)
            # if it sees '/', it looks for and returns index.htm (or index.html)
            # 2 ways to call the Parent... name the parent, or super the child...
            SimpleHTTPRequestHandler.do_GET(self)
            logging.debug("MAIN:do_GET Path: %s\nHeaders:\n%sRequestLine:%s\nMAIN:do_GET:Done.", str(self.path), str(self.headers), self.requestline)
            #super(PiGS_HTTPRequestHandler, self).do_GET()

    def do_POST(self):  # this receives URLs like POST /UIsettings or
        
        global dPiGS
        
        logging.debug("MAIN:POST: Path: %s Headers: %s", str(self.path), str(self.headers))

        # Break url into scheme, netloc, path, params, query,fraqment
        # then parse the query parameters into a dictionary
        o=urlparse(self.path)
        qs=parse_qs(o.query)

        # get and decode the body json
        content_length = int(self.headers['Content-Length'])
        body = self.rfile.read(content_length)
        
        # generic decode for error return/debug, may be re-decoded for specific command
        bj = json.loads( body.decode() )
        
        # Now process the individual Post commands
        
        if ("UIsettings" in self.path) :  # POST settings
           
            # Settings returned from the UI
            # ON Save Settings, lineup or icons may have changed we need to abort any build in progress
            # If a schedule build was in progress, UI Save would kill the build b/c the token would change.
            
            #Todo disable UI button until this command returns???
            
            if not SDResourceLock.acquire(False) :  # if we get the lock, we can do the settings
                
                logging.warning("MAIN:POST:UIsettings: PiGS is busy building schedule.  Requesting Abort...")
                
                BuildAbortEvent.set()   #signal Linear Schedule Builder to abort
        
                time.sleep(1)    # give thread time to end
                
                waitAttempts = 5 #seconds
                while waitAttempts and not SDResourceLock.acquire(False) :
                    time.sleep(1)
                    waitAttempts -= 1

                BuildAbortEvent.clear() # ensure flag is clear after abort
                
                if waitAttempts == 0 :
                    logging.warning("MAIN:POST:UIsettings: Timeout on Abort. PiGS could not save Settings!")
                    self.send_error(ServerErrorCode, message="PiGS is busy building the Schedule")
                    #We never got the lock so I dont have to release it, but better to be safe
                    #if SDResourceLock.locked() :
                    #    SDResourceLock.release()
                    return
                
            # We have the Lock, we can make SD calls, or Queue commands
            # If this function returns before Queuing the Service build, must do a SDResourceLock.release()
            
            #--------------------------------------------------------------
            # Get the settings from the UI message block, json to dict conversion
            PigsUIreturned = json.loads( body.decode() )
            logging.info("MAIN:POST:UIsettings len: %s", content_length )
            self.send_response(200)
            self.end_headers()

            #--------------------------------------------------------------
            # I could just replace the whole dPiGS dict with the received UI json.
            #   but maybe there are some keys we want to preserve on the server side
            # I do this with keys in pigsPreserveSettings.  See notes where dict is defined.
            # Copy some keys from current pigs settings to received settings
            # PigsPreserveSettings is just a list of keys only
            # Update the returned keys with current keys in case a few special keys changed while UI opened
            
            for key in PigsPreserveSettings :  # Use local definition structure
               PigsUIreturned[key] = dPiGS[key] 

            # Do I need to sanity check the data ???
            # should I check the data before saving?
            # should I make the changes on the Temp settings like do the SD stuff first?
            # Just save what the User configured, if there is a fault at least we saved users work
            # when they restart, refresh UI and click save settings all should be well

            #--------------------------------------------------------------
            # Recursively copy the UI data into our main config dict
            dPiGS = copy.deepcopy(PigsUIreturned)
            
            # This saves 5 backps as <name>N.json, where N = 1..5
            SaveSettingsFile(dPiGS, CONFIG_FILE)

            #--------------------------------------------------------------
            #https://docs.python.org/3/library/logging.html
            #See the heading 'Module Level Functions'
            #Get for a reference to the default root level logger
            logger = logging.getLogger()
            logger.setLevel( dPiGS['LoggingLevel'] )
            
            
            #--------------------------------------------------------------
            # Process PiGS DNS Server changes
            global IPaddr
            
            # Starting DNS
            if dPiGS['PigsDoesDNS'] and not PigsDnsAlive() :  # do we need to start DNS?
                if dPiGS['PigsDelayDNS'] :
                    logging.info("MAIN:UIsettings: DNS will be started after Schedule is built...")
                else :
                    logging.info("MAIN:UIsettings: Starting DNS...")
                    PigsDnsStart(dPiGS['Server_IP'], IPaddr, dPiGS['DnsHosts'])
            
            # Stopping DNS
            if not dPiGS['PigsDoesDNS'] and PigsDnsAlive() :
                logging.info("MAIN:UIsettings: Stopping DNS - waiting for thread to end...")
                dnsThread = PigsDnsStop()
                dnsThread.join(2.0)  # wait up to 2 seconds for DNS to Stop (should stop in less than 1 sec)
                logging.info("MAIN:UIsettings: DNS Stopped... Alive=%s", dnsThread.is_alive())

            #--------------------------------------------------------------
            # Process Lineup changes on Schedules direct
            # We need to make SD match what user configured by deleting and adding lineups
            
            # create a list of the current lineup names only
            newLineups = []
            for L in dPiGS['MyLineups'] :
                newLineups.append( L['LineupName'] )
            logging.info("MAIN:POST:UIsettings: Queuing BuildAddRemove...")
            
            # send command and Params as dictionary
            BuilderQueue.put( {'Command':BuildAddRemove, 'Params':newLineups} )
            
            # After making UI changes, rebuild services list (list of channel favorites)
            # DVR+ user will need to do a full guide refresh so it requests the new lineup services
            
            # We already have the SDResourceLock so we can just add the Build Cmd to the Queue
            # We must explicitly do this build here
            logging.info("MAIN:POST:UIsettings: Queuing BuildServices...")
            BuilderQueue.put( {'Command':BuildServices} )
            
            # And we can schedule a SD Schedule build for 'now'
            logging.info("MAIN:POST:UIsettings: Queuing BuildSchedule...")
            bts = SetDvrBuildTimer(1)  # set timer for 1 second
            logging.info("MAIN:POST:UIsettings: Build time scheduled for %s", bts)
            
            # Done processing UI settings

        # User hits 'Verify Login' to create a sha1 password hash, Test login by getting SD token
        if ("sdlogin" in self.path) :

            sha1pwd = GetSha1Password(bj['password'])
            logging.info("MAIN:POST:SDLogin Attempting GetToken")

            if not SDResourceLock.acquire(False) :  # if we get the lock, we can do the sdlogin
                logging.warning("MAIN:POST:SDLogin Can't login.  PiGS is busy rebuilding schedule!")
                self.send_error(ServerErrorCode, message="PiGS is busy.  Please Retry.")
                return  # didnt get the lock, dont have to release anything

            # Builder will signal us when Login data is ready
            LoginReadyEvent = threading.Event() 
            LoginReadyEvent.clear()

            # create a build command
            lgcmd = copy.copy(BldCmd)
            lgcmd['Command'] = BuildLogin
            lgcmd['Params']  = {'un':bj['username'],'pw':sha1pwd }
            lgcmd['Event']   = LoginReadyEvent
            
            logging.info("MAIN:POST:SDLogin: Queuing BuildLogin & waiting for Result...")
            BuilderQueue.put( lgcmd )

            # The PiGS webserver is busy during this wait and cant respond to other commands
            LoginReady = LoginReadyEvent.wait(10) # wait here for up to 10 seconds
            
            if LoginReady :
                logging.info("MAIN:POST:SDLogin: BuildLogin is ready.")
                d = lgcmd['Result']   # get reference to the result dictionary 
                respcode  = d['code']
                rsj       = d['errj']
            else :
                logging.warning("MAIN:POST:SDLogin: Timeout waiting for BuildLogin Result.")
                respcode = 503
                rsj  = None
           
            # if it works, the username/sha1 combo is good and we can save in the config file
            if respcode == 200 :
                # update the config dictionary
                logging.info("MAIN:POST:SDLogin: SD Login Successful!")
                dPiGS['sdusername'] = bj['username']
                dPiGS['sdsha1pwd']  = sha1pwd
                SaveSettingsFile(dPiGS, CONFIG_FILE)     # Save setting file to disk
            else :
                logging.warning("MAIN:POST:SDLogin: code(%s) %s", respcode, rsj)

            self.send_response(respcode)
            self.end_headers()

    def _do_unsupported(self):
        logging.warning("MAIN:Unsupported Command: do_ %s", self.command)
        logging.warning("MAIN: Path: %s\nHeaders:%s", str(self.path), str(self.headers))
        self.send_response(400)
        self.end_headers()
        response = BytesIO()
        response.write(b'This is an unsupported request. ')
        
    #https://stackoverflow.com/questions/43156023/what-is-http-host-header
    #https://bugs.chromium.org/p/chromium/issues/detail?id=521009
    #https://superuser.com/questions/1398098/how-can-i-get-google-chrome-to-go-to-an-internal-page-without-performing-an-exte
    #https://docs.microsoft.com/en-us/troubleshoot/browsers/one-word-intranet-sites
    def do_HEAD(self):
        logging.info("MAIN:do_Head. Passing to parent.")
        SimpleHTTPRequestHandler.do_HEAD(self)
        logging.debug("MAIN:do_HEAD Path: %s\nHeaders:\n%sRequestLine:%s\nMAIN:do_HEAD:Done.", str(self.path), str(self.headers), self.requestline)
    def do_DELETE(self):
        self._do_unsupported()
    def do_PATCH(self):
        self._do_unsupported()
    def do_OPTIONS(self):
        self._do_unsupported()
    def do_PUT(self):
        self._do_unsupported()

#---------------------------------------------------------------------------------------
# Function to get my IP for display and wait for network to boot
# When starting PiGS at boot, the network is not always ready and
# returns: [Errno 101] Network is unreachable
# We will sleep a few times while waiting for network to boot
# 30 seconds max wait = 30 attempts, 1 sec wait each time
# 
# None is the return value if cant get IP

import socket

def WaitForNetwork(NetworkAttempts = 30):

    for i in range(NetworkAttempts, -1,-1) : # start, stop 1 before, step
    
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        try:
            # doesn't even have to be reachable
            s.connect(('10.255.255.255', 1))
            IP = s.getsockname()[0]
        except Exception as err:
            logging.debug("  get_ip = %s", err )
            IP = None
        finally:
            s.close()
    
        if IP :        # If we get real IP, break out of wait loop
            break
        
        logging.info("  Waiting for network %d", i )
        time.sleep(1)      # take a nap while the system boots the Network 

    #time.sleep(1)      # take a nap while the system boots the Network 

    return IP

#---------------------------------------------------------------------------------------
# this helps put the log file and config files in the pigs.py directory
# even when app is launched when current working dir is '/'

def PigsPath(fn) :
    
    pigspypath = os.path.realpath(__file__)
    
        # separate file path/name from its extention
    #root, ext  = path.splitext(configFile)
    # separate path from 'pigs.py'
    head, tail = os.path.split(pigspypath)

    return os.path.join(head, fn)



##########################################################################
# OK, here we go... initialize and start things going

print("\nStarting...")

# Some print()'s are used so something appears when logging to a file

#https://docs.python.org/3/howto/logging.html
#https://docs.python.org/3/library/logging.html#levels
#https://docs.python.org/3/library/logging.html#logging.basicConfig
#https://docs.python.org/3/library/logging.html#module-level-functions
#https://docs.python.org/3/library/logging.html
#See the heading Module Level Functions

#Level      When it’s used
#NOTSET      0
#DEBUG      10 Detailed information, typically of interest only when diagnosing problems.
#INFO       20 Confirmation that things are working as expected.
#WARNING    30 An indication that something unexpected happened, or indicative of some problem in the near future (e.g. ‘disk space low’). The software is still working as expected.
#ERROR      40 Due to a more serious problem, the software has not been able to perform some function.
#CRITICAL   50 A serious error, indicating that the program itself may be unable to continue running.
#The default level is WARNING

# Open CONFIG_FILE JSON file, load settings into a Dict
# Cant do any logging before the config file since we get the log file name from there
#https://stackoverflow.com/questions/9823936/python-how-do-i-know-what-type-of-exception-occurred
#https://docs.python.org/2/library/exceptions.html#exception-hierarchy

bDelayedLogging = False

print("PiGS Working Directory on launch:" , os.getcwd())

# V0.99 Change PiGS working directory to where program started
head, tail = os.path.split( os.path.realpath(__file__) )
os.chdir(head)

#====================================================================================
# Test for icons folders, create if not found.
# the mode params do not seem to work for mkdir

if not os.path.exists('myicons'):
    os.mkdir('myicons')  #, mode=0o644, *, dir_fd=None)
    os.chmod('myicons', mode=0o777)
if not os.path.exists('cmicons'):
    os.mkdir('cmicons')  #, mode=0o644, *, dir_fd=None)
    os.chmod('cmicons', mode=0o777)
if not os.path.exists('sdicons'):
    os.mkdir('sdicons')  #, mode=0o644, *, dir_fd=None)
    os.chmod('sdicons', mode=0o777)


#====================================================================================

cfgf = PigsPath(CONFIG_FILE)
    
try:
    f = open(cfgf)          # 'pigs.json'
    dPiGS = json.load(f)   # returns JSON object as a dictionary 
    f.close()              # Closing file

# JSONDecodeError exception isnt in Python 3.4 and earlier, dont really need it
#except json.JSONDecodeError:  
#    print("Error decoding config file.", CONFIG_FILE)
#    sys.exit(5)

except FileNotFoundError:
    print("Config file %s not found, creating...", cfgf)
    dPiGS = copy.copy(PigsConfigDict)
    with open(cfgf, "w") as outfile: 
        json.dump(dPiGS, outfile, indent=2)
        outfile.close()
        
    # need to delay logging this b/c logging not enabled yet - log file name in config file
    bDelayedLogging = True
    # We can contine, but this is pretty serious unless it is the very first run
    # I guess its not too serious...  if its gone, you just have to reconfigure.
    
except Exception as ex:
    template = "An exception of type {0} occurred. Arguments:\n{1!r}"
    message = template.format(type(ex).__name__, ex.args)

# update the version number in the config file in case the app version has changed.
dPiGS['PiGS_Version'] = PIGS_VERSION


#====================================================================================
# Auto-Upgrade pigs.json with missing keys due to version upgrades

if 'DnsHosts' not in dPiGS :        # PiGS acts as a DNS Server for DVR+.  Forwards some requests.
    dPiGS["DnsHosts"] = HostsDict   # Default new key to false to not mess up current users

if 'IconType' not in dPiGS :        # Icon type 1=SD, 2=CM
    dPiGS["IconType"] = 2 if dPiGS["UseCMImages"] else 1
    dPiGS.pop('UseCMImages',0)      # remove old key nolonger used default 0 prevent exception
    
if 'PigsDoesDNS' not in dPiGS :     # PiGS acts as a DNS Server for DVR+.  Forwards some requests.
    dPiGS["PigsDoesDNS"] = False    # Default new key to false to not mess up current users
if 'PigsDelayDNS' not in dPiGS :    # PiGS acts as a DNS Server for DVR+.  Forwards some requests.
    dPiGS["PigsDelayDNS"] = True    # Default new key to True, start DNS after Schedule is built
if 'Server_IP' not in dPiGS :
    dPiGS["Server_IP"] = '0.0.0.0'  # defalt "0.0.0.0"
if 'Server_Port' not in dPiGS :
    dPiGS["Server_Port"] = 8000     # defalt 8000.  use cmdline to override to 80

if 'RebuildSecs' not in dPiGS :
    dPiGS["RebuildSecs"] = 82800  # 23 hours = 23h*60*60 Next Build Time in seconds
if 'MGRnewSched' not in dPiGS :
    dPiGS["MGRnewSched"] = True    # Manual Guide Refresh forces Rebuild of SD Schedule
if 'AllowAnyZip' not in dPiGS :
    dPiGS["AllowAnyZip"] = True    # allow any text in Zip box.  False = USA 5 digit, numbers only
if 'SaveFsSDIcons' not in dPiGS :
    dPiGS["SaveFsSDIcons"] = False # keeps a local copy of the resized SD Icons
if 'SDImSize' not in dPiGS :
    dPiGS["SDImSize"] = [72,54]    # 360w x 270 / 5   = 72 x 54
if 'SDImResample' not in dPiGS :
    dPiGS["SDImResample"] = 3      # Deault to BICUBIC.   Pillow Lib .resize() Resample type 0..5
if 'SDImReduceGap' not in dPiGS :
    dPiGS["SDImReduceGap"] = None  # Only in Pillow 7.0 - Reducing Gap
if 'CacheSDIcons' not in dPiGS :
    dPiGS["CacheSDIcons"] = True  # keeps a local copy of the resized SD Icons
if 'CacheCMIcons' not in dPiGS :
    dPiGS["CacheCMIcons"] = True   # keeps a local copy of the CM / Rovi Icons
if 'DefaultIcon' not in dPiGS :
    dPiGS["DefaultIcon"] = "pinkpig.png"
if 'SdIconTheme' not in dPiGS :
    dPiGS["SdIconTheme"] = "dark" # choices are dark, gray, light, white
if 'TimeZones' not in dPiGS :
    dPiGS['TimeZones'] = copy.deepcopy(PigsConfigDict['TimeZones'])
        
if dPiGS['TimeZones'][0]['Offset'] == "???" :  # ??? if flag to tell us to init
    # Set the timezone object to the current time zone on each boot
    # ToDo what happens if user changes timezone while pigs is running
    s,offset = LocalTimeStr( datetime.utcnow() )
    dPiGS['TimeZones'][0]['Offset'] = str(offset)

#====================================================================================
# Check for command line args.
# Pass in Port and maybe logging level to use config for logging level
# Port 80 is 'protected' so PiGS must be started with the sudo command
#   sudo python3.7 pigs.py 80
#   sudo python3 pigs.py 80
# to allow the app to 'bind' to port 80.  No error checking is done here

# Get the port number if provided, otherwise default to 8000
if len(sys.argv) > 1:
   #logging.info("MAIN: SYS args name=%s Port=%s", sys.argv[0], sys.argv[1])
   port = int(sys.argv[1])
else :
   #port = SERVER_PORT
    port = dPiGS["Server_Port"]
    
    
#====================================================================================
# Get logging settings from config file, then Enable Logging

logFile  = dPiGS.get('LoggingFile',  PIGS_LOG_FILE)  # if logFile='' it will log to console
logLevel = dPiGS.get('LoggingLevel', logging.INFO)
logf     = PigsPath(logFile)

# if running from the Dev environment, dont log to file so as not to
# conflict with the port 80 process if it is running.
# running both at the same time could have side effects with the config file also

#if port != 80 :     # if not port 80, remove the filename to log to console
if port == dPiGS["Server_Port"] :  # if 8000, remove the filename to log to console
    logf = ""          

# replaced this code with the RotatingFileHandler
#https://stackoverflow.com/questions/24505145/how-to-limit-log-file-size-in-python
#https://gist.github.com/amamaenko/653743
# initialize logging before making calls to any modules that use logging or weird(nothing) thing happen

logger = logging.getLogger()

# add a Rotating File Handler to keep log file size in check
if logf :
    handler = logging.handlers.RotatingFileHandler(logf, maxBytes=MAX_LOGFILE_BYTES, \
                                                   backupCount=LOGFILE_BACKUPS)
    logger.addHandler(handler)
    logger.setLevel( logLevel )
else :
    logger = logging.basicConfig(filename=logf,level=logLevel)

# Also send Info logs to the console
# docs.python.org/2.4/lib/multiple-destinations.html
console = logging.StreamHandler()
console.setLevel(logging.INFO)

#formatter =logging.Formatter('%(name-12s: %levelname)-8s %(message)s')
#console.setformatter(formatter)
#if port != 80 :     # again, if command line start, log to console also
if port != dPiGS["Server_Port"] :  # again, if command line start, log to console also
    logging.getLogger("").addHandler(console)

if logf : print("Logging to file:", logf)
print("Logging Level is", logLevel)

# set global flag to over-ride annoying HTTP log.info prints
bLogHttp = dPiGS.get('LogHttpReq')

PigsStartTime = datetime.utcnow()
lst, lz = LocalTimeStr( PigsStartTime )
logging.info("MAIN: PiGS started at %s", lst )
logging.warning("Logging Level is %s", logLevel)
logging.warning("MAIN: PiGS Version %s", dPiGS['PiGS_Version'] )
logging.info("MAIN: Current Working Directory %s" , os.getcwd())

# Had to delay logging this because the logfile did not yet exist when error occurred
if bDelayedLogging :
    logging.warning("MAIN: Config file %s not found, created!", cfgf)

#====================================================================================
#
logging.info("MAIN: Platform Info")
logging.info("  MaxSize(%s), Is64-bit(%s)", hex(sys.maxsize), sys.maxsize > 2**32 )

try :
    import platform
    logging.info("  Achritec: %s", platform.architecture() )
    logging.info("  System  : %s", platform.system() )
    logging.info("  Machine : %s", platform.machine() )
    logging.info("  Platform: %s", platform.platform() )
    logging.info("  Python I: %s", platform.python_implementation() )
    logging.info("  Python V: %s", platform.python_version() )
except :
    logging.warning("  Platform library not available")
    
#Python Version as a Float for comparisons
PythonVer = VersionStringToFloat(platform.python_version())

logging.info("MAIN: http.server Info")
try :
    logging.info("    server_version: %s", PiGS_HTTPRequestHandler.server_version )
    logging.info("       sys_version: %s", PiGS_HTTPRequestHandler.sys_version  )
    logging.info("  protocol_version: %s", PiGS_HTTPRequestHandler.protocol_version)
except :
    logging.warning("MAIN: Error getting http.server version info.")

#====================================================================================
# Wait for Network Ready and Print the Pigs IP address.

IPaddr = WaitForNetwork()
logging.info("MAIN: PiGS Discovered IP = %s", IPaddr )
logging.info("MAIN: PiGS Serving on Port %d", port )

dPiGS['PiGS_IP'] = IPaddr  # save PiGS_IP in config to display on UI

# manual over-ride of listen IP address for port 80 & port 53
if dPiGS['Server_IP'] != '0.0.0.0' :
    
    dPiGS['PiGS_IP'] = dPiGS['Server_IP']  # UI
    
    # Print a warning if my discovered IP does not match the set IP
    # This can happen if a device have multiple Ethernet interfaces like a Wired and a Wifi
    # This message lets user know which is bound
    if dPiGS['Server_IP'] != IPaddr :
        logging.warning("MAIN: ******* Server_IP(%s) does not match discovered IP(%s) ******* ", dPiGS['Server_IP'],IPaddr )


#====================================================================================
# Start the Builder Thread
logging.info("MAIN: Starting BuilderThread")
BuilderThread = threading.Thread(target=Builder, name='Builder', daemon=True )
#BuilderThread = threading.Thread(target=Builder, name='Builder', daemon=True, args=(dPiGS, PiGS_SERVICEID, BuildAbortEvent,) )
BuilderThread.start()
logging.info("MAIN: BuilderThread.isalive(): %s", BuilderThread.is_alive())

#====================================================================================
# Start the DNS Server Thread
 
#logging.info("MAIN: DNSThread.isAlive(): %s", PigsDnsThread.is_alive())

#Todo - this should be moved to after sched builder completes
if dPiGS["PigsDoesDNS"] :
    if dPiGS["PigsDelayDNS"]:
        logging.info("MAIN: Delaying DNS Start until Schedule is Built.")
    else :
        PigsDnsStart(dPiGS['Server_IP'], IPaddr, dPiGS['DnsHosts'])

#====================================================================================
# Build/Load the Service details so they are available to the Schedule builder.
# On very first run there is no user name so skip service build.
# It will get built when user does a Save Settings in the UI
# If we never got an IP address, dont attempt to build the Service or Schedule lists

if IPaddr and dPiGS['sdusername'] :

    if SDResourceLock.acquire(False) :  # if we got the lock, Queue Services build
        
        # A thread delay so it looks like the build starts after initialization
        BuilderQueue.put( {'Command':BuildDelay} )
        
        # new for 1.01, verify SD account
        # Verify SD account lineups match what user configured by deleting and adding lineups
        # create a list of the current lineup names only
        newLineups = []
        for L in dPiGS['MyLineups'] :
            newLineups.append( L['LineupName'] )
        logging.info("MAIN: Queuing BuildAddRemove to verify Lineups...")
        # send command and Params as dictionary
        BuilderQueue.put( {'Command':BuildAddRemove, 'Params':newLineups} )

        # Build the Service list (Lineup list)
        logging.info("MAIN: Queuing BuildServices...")
        BuilderQueue.put( {'Command':BuildServices} )
        
        # Schedule the build of the SD Schedule 2 seconds from now
        logging.info("MAIN: Starting 2 second BuilderTimer")
        SetDvrBuildTimer(2.0)
        
    else :
        logging.warning("MAIN: Could not get SDResourceLock() to Build Service & Schedule!")
else :
    logging.warning("MAIN: No SD username, can't Build Schedule! Configure at %s", 'http://'+IPaddr)


#====================================================================================
# Construct and start the webserver

hostIP = dPiGS["Server_IP"]    # defualt "0.0.0.0",
  
logging.info('MAIN: HTTPServer binding to IP=%s',hostIP)

# create/bind instance of a web server object with my custom handler
Handler = PiGS_HTTPRequestHandler

try :
    httpd   = HTTPServer((hostIP, port), Handler)
except:
    logging.warning("MAIN: Could not bind to port %s", port )
    logging.warning("MAIN: Another PiGS runninng?  Try 'sudo lsof -i :80' for pid" )

logging.info("MAIN: Waiting for DVR+ requests on Port %d!",port)

# functions called by server are not thread safe so this doesnt work.
# httpd   = ThreadingHTTPServer((hostIP, port), Handler)

# And start it... this call does not end, 
# But allow for clean exit from keyboard with Ctrl-C
try:
    httpd.serve_forever()
except KeyboardInterrupt:
    pass

httpd.server_close()

logging.info('MAIN: Kbd Interrupt - Stopping httpd...')

# Save Settings on exit to preserve build times 
logging.info('MAIN: Saving User Settings...')
SaveSettingsFile(dPiGS, cfgf)

logging.info("MAIN: PiGS done at utc %s\n", datetime.utcnow())

