# sd.py
# functions to interface with or manipulate Schedule Direct
# Copyright June 2020, Timothy Goihl
#
# This module was developed for the '2014' SD API
# https://github.com/SchedulesDirect/JSON-Service/wiki/API-20141201
#

PIGS_USER_AGENT = 'PiGS/V1.0'
import json
import requests
import logging
import datetime

#https://dateutil.readthedocs.io/en/stable/
SchedulesDirectURL   = "http://json.schedulesdirect.org"
TokenPostURL    = "http://json.schedulesdirect.org/20141201/token"
StatusGetURL    = "http://json.schedulesdirect.org/20141201/status"
AvailableGetURL = "http://json.schedulesdirect.org/20141201/available"
HeadendsGetURL  = "http://json.schedulesdirect.org/20141201/headends" #/country=USA&postalcode=60030

PreviewLineupsGetURL = "http://json.schedulesdirect.org/20141201/lineups/preview/"  #/lineupcode
PutUserLineupsURL    = "http://json.schedulesdirect.org/20141201/lineups/" #/lineup code
GetUserLineupsGetURL = "http://json.schedulesdirect.org/20141201/lineups"
DelUserLineupDelURL  = "http://json.schedulesdirect.org/20141201/lineups/"  #lineup code

LineupsMapGetURL      = "http://json.schedulesdirect.org/20141201/lineups/" #+lineup code
SchedulesPostURL      = "http://json.schedulesdirect.org/20141201/schedules"
SchedulesMD5PostURL   = "http://json.schedulesdirect.org/20141201/schedules/md5"
ProgramDetailsPostURL = "http://json.schedulesdirect.org/20141201/programs"
ProgramGenerucPostURL = "http://json.schedulesdirect.org/20141201/metadata/description"

#Useful Test website that echos the Post
# response = requests.post("http://httpbin.org/post", json={"username":"sdusername","password":'a sha1 has goes here'})

#Programming references
#https://pynative.com/python-post-json-using-requests-library/
#https://requests.readthedocs.io/en/master/user/quickstart/#json-response-content
#https://requests.readthedocs.io/en/master/user/advanced/#request-and-response-objects

#Chunked encoding
#https://requests.readthedocs.io/en/master/user/advanced/#chunk-encoded-requests

#==============================================================================
# Token Handling
#
# Utility function to Generate Sha1 password encoding required by SD
# Python shell usage:
#   from sd import GetSha1Password
#   GetSha1Password('mySDpassword')

import hashlib

def GetSha1Password(pwd) :
    #str = "mypassword"
    sha1pw = hashlib.sha1( pwd.encode() ).hexdigest()
    return sha1pw

# global gToken for use by other SD APIs when they need it

gToken = None

# Returns http error status 200,400, etc and the json so we can get error details
# If a token is passed in, just set that instead of asking for one
# this is for devlopent so that if using Postman on PC we are using the same token
# the token is saved in the config file and has to be generated manually

def SDGetToken(username, sha1pwd, token=None ) :

    global gToken

    # If a token is passed in, set gToken and then exit with a http success code 200
    if (token) :        
        gToken = token
        return 200, gToken

    gToken = None
    
    newHeaders = {'User-Agent':PIGS_USER_AGENT}

    #send the request with the un/pw in the body    
    r = requests.post(TokenPostURL, json={"username":username,"password":sha1pwd}, headers=newHeaders)

    if r.status_code >= 500 :
        logging.warning("SD: SDGetToken http(%d)", r.status_code)    
        return r.status_code, 0, None
    
    rj = r.json()

    # Bad or No username/password is a 403 error.
    if r.status_code != 200 :
        logging.warning("SD: SDGetToken http(%d) SDcode(%d). Message: %s", r.status_code, rj.get('code'), rj.get('message') )    
        return r.status_code, 0, rj
    
    if rj['code'] == 0:
        gToken = rj['token']
        logging.info("SD:SDGetToken codes: %d, %d, Token:%s", r.status_code, rj['code'], rj['token'])
    else :
        logging.warning("SD:SDGetToken codes: %d, %d, SD-MSG:%s", r.status_code, rj['code'], rj['message'])
        
    #return token or None
    return r.status_code, rj['code'], rj['message']


#==============================================================================
# Process Lineup changes on Schedules direct
# We need to make SD match  what user configured by deleting and/or adding lineups
#
# Get list of lineups from SD into  sdLuList
# for lu in sdLuList, if lu not in MyLineup, delete SD list
# get sdLuList
# for ea lu in MyLineups, if not in sdLuList, ADD it
#
# Caller must get token since we dont have un/pwd

def SDAddDeleteLineups( UiLineupList ) :
    
    # Get current account lineups

    httpcode, dLineups = GetLineups()
    
    logging.info("SD:SDAddDeleteLineups: PiGS Lineups %s ", UiLineupList)

    newHeaders = {'User-Agent':PIGS_USER_AGENT}

    # 400 seems to be No lineups in account
    if httpcode != 200 :
        logging.warning('SD:SDAddDel: Error getting SD Lineups (Maybe no lineups) Code:%d, %s', httpcode, dLineups)

    # We have to continue b/c we may be adding the first lineup

    logging.debug("SD:Current Lineups in Account")
    if dLineups :   # need the if because there is a key lookup on the next line
        for sdl in dLineups['lineups'] :  
            logging.debug("SD:   %s", sdl['lineup'] )

    logging.debug("SD:Current User Lineups") 

    for uil in UiLineupList :  # this for() works on an empty list
        logging.debug("SD:   %s", uil )

    # First do Deletes - if item in the SD Account list is not in the UI list 
    if dLineups :   # need the if because there is a key lookup on the next line
        for sdl in dLineups['lineups'] :
            if sdl['lineup'] not in UiLineupList :
                logging.info('SD: Deleting SD Lineup: %s', sdl['lineup'] )
                httpcode, sdcode = DeleteLineup(sdl['lineup'])
    
    # Next do Add's
    # For each lu in UiList, if not in SA account, Add it
    
    for uil in UiLineupList :
        #I want to do this >>> if uil not in dLineups['lineups'] :
        #   but we need to interate sdl['lineup'][] 'name'
        #   I'm sure Python can do this, I just dont know how
        if not dLineups : # empty SD list, just add it
            logging.info('SD: Adding SD Lineup: %s', uil)
            httpcode, sdcode = AddLineup(uil)
        else : # search existing SD lists
            for sdl in dLineups['lineups'] :
                if uil == sdl['lineup'] :
                    break  # found, break out, its already there, do next 
            if uil != sdl['lineup'] :  # if the loop ends w/o a match
                logging.info('SD: Adding SD Lineup: %s', uil)
                httpcode, sdcode = AddLineup(uil)


    # now print what the account has in it for testing
    logging.debug("SD:New Lineups in Account:")
    httpcode, dLineups = GetLineups()
    if (httpcode == 200) and dLineups :
        for sdl in dLineups['lineups'] :
            logging.debug("SD:   %s", sdl['lineup'] )
        
    return httpcode

# I want to do this >> if uil not in dLineups['lineups'] :
#https://stackoverflow.com/questions/18210436/python-custom-class-operator-overloading
#https://realpython.com/operator-function-overloading/
#https://stackoverflow.com/questions/38542543/functionality-of-python-in-vs-contains/38542777
#https://stackoverflow.com/questions/1964934/what-does-contains-do-what-can-call-contains-function
        #if uil not in dLineups['lineups'] :
        # Cant use if uil not in dLinupes[] bc we need to compare to sdl['lineup']
#{'ID': 1, 'alist': [{'lineup': '123'}, {'lineup': 'abc'}]}




#==============================================================================
#
def SDGetStatus() :
    
    global gToken
    
    logging.debug("SD:GetStatus ")
    
    # header needs the token
    newHeaders = {'User-Agent':PIGS_USER_AGENT, 'Content-type':'text/plain','Accept':'gzip, deflate','token':gToken}

    # send a GET request
    r1 = requests.get(StatusGetURL, headers=newHeaders)
    
    # Handle success first
    if r1.status_code == 200 :
        rj = r1.json()
        return r1.status_code, rj
    
    if r1.status_code == 400 :  # prob no lineups to get
        rj = r1.json()
        logging.warning("SD:GetStatus http(%d) Message: %s", r1.status_code, rj )    
        return r1.status_code, None   # no lineups to send to user
    
    if r1.status_code == 500 :
        logging.warning("SD:GetStatus http(%d) Message: %s", r1.status_code, r1.text )    
        return r1.status_code, None
    
    # some other error
    #rj = r1.json()
    
    return r1.status_code, None  #rj.get('code')



#==============================================================================
#
def GetLineups() :
    
    global gToken
    
    logging.debug("SD:GetLineups ")
    
    # header need the token
    newHeaders = {'User-Agent':PIGS_USER_AGENT, 'Content-type':'text/plain','Accept':'gzip, deflate','token':gToken}
    #myparams = {'country': 'USA', 'postalcode': zipcode}

    # send a GET request
    r1 = requests.get(GetUserLineupsGetURL, headers=newHeaders)
    
    # Handle success first
    if r1.status_code == 200 :
        rj = r1.json()
        return r1.status_code, rj
    
    if r1.status_code == 400 :  # prob no lineups to get
        rj = r1.json()
        logging.warning("SD:GetLineups http(%d) Message: %s", r1.status_code, rj )    
        return r1.status_code, None   # no lineups to send to user
    
    if r1.status_code == 500 :
        logging.warning("SD:GetLineups http(%d) Message: %s", r1.status_code, r1.text )    
        return r1.status_code, None
    
    # some other error
    #rj = r1.json()
    
    return r1.status_code, None  #rj.get('code')

#==============================================================================
#
def AddLineup(luname) :  # Add line upt to user account
    
    global gToken
    
    logging.debug("SD:AddLineup: %s", luname)

    # header need the token
    newHeaders = {'User-Agent':PIGS_USER_AGENT, 'Content-type':'text/plain','Accept':'gzip, deflate','token':gToken}
    #myparams = {'country': 'USA', 'postalcode': zipcode}

    # send a PUT request to add the lineup
    r1 = requests.put(PutUserLineupsURL + luname, headers=newHeaders)
    
    # Handle success first
    if r1.status_code == 200 :
        rj = r1.json()
        logging.info("SD:AddLineup: %d, %d %s", r1.status_code, rj.get('code'), rj.get('message') )
        logging.info("SD:AddLineup: %d changes remaining.", rj.get('changesRemaining') )
        
        return r1.status_code, rj.get('code')  #0=ok, #2105 invalide lineup
    
    if r1.status_code == 500 :
        logging.warning("SD:AddLineup http(%d) Message: %s", r1.status_code, r1.text )    
        return r1.status_code, None
    
    # some other error
    #rj = r1.json()
    
    return r1.status_code, None  #rj.get('code')

#==============================================================================
#
def DeleteLineup(luname) :
    
    global gToken
    
    logging.debug("SD:DelLineup Lineup: %s", luname)
    
    # header need the token
    newHeaders = {'User-Agent':PIGS_USER_AGENT, 'Content-type':'text/plain','Accept':'gzip, deflate','token':gToken}
    #myparams = {'country': 'USA', 'postalcode': zipcode}

    # send a Delete request
    r1 = requests.delete(DelUserLineupDelURL + luname, headers=newHeaders)
    
    # Handle success first
    if r1.status_code == 200 :
        rj = r1.json()
        logging.info("SD:DelLineup http:%d, sd:%d %s", r1.status_code, rj.get('code'), rj.get('message') )
        #logging.info("SD:AddLineup: %d, %d %s", r1.status_code, rj.get('code'), rj.get('message') )
        logging.info("SD:DelLineup: %d changes remaining.", rj.get('changesRemaining') )
        return r1.status_code, rj.get('code')
    
    if r1.status_code == 500 :
        logging.warning("SD:DelLineup http(%d) Message: %s", r1.status_code, r1.text )    
        return r1.status_code, None
    
    # some other error
    #rj = r1.json()
    
    return r1.status_code, None  #rj.get('code')

#==============================================================================

#==============================================================================
#https://requests.readthedocs.io/en/master/user/quickstart/
#>>> payload = {'key1': 'value1', 'key2': 'value2'}
#>>> r = requests.get('https://httpbin.org/get', params=payload)

def SDGetStationMapping(LineupName) :

    global gToken
    
    logging.debug("SD:SDGetStationMapping Lineup: %s", LineupName)

    # header need the token
    newHeaders = {'User-Agent':PIGS_USER_AGENT, 'Content-type':'text/plain','Accept':'gzip, deflate','token':gToken}
    #myparams = {'country': 'USA', 'postalcode': zipcode}

    # send a GET request
    r1 = requests.get(LineupsMapGetURL + LineupName, headers=newHeaders)
    
    # Handle success first
    if r1.status_code == 200 :
        rj = r1.json()
        return r1.status_code, rj
    
    if r1.status_code == 500 :
        logging.warning("SD:SDGetStationMapping http(%d) Message: %s", r1.status_code, r1.text )    
        return r1.status_code, None
    
    # some other error
    #rj = r1.json()
    
    return r1.status_code, None   #rj.get('code')


#==============================================================================
# 2-stage call
# 1st - using postalcode, get the headends, search for Antenna with matching Zip
# 2nd - ask for a lineup preview. 
#

def SDGetLineupPreview(zipcode) :

    global gToken
    
    
    # ensure canadian zipcodes are uppdercase to match SD's Headend results
    zipcode = zipcode.upper()
    
    # header need the token
    newHeaders = {'User-Agent':PIGS_USER_AGENT, 'Content-type':'text/plain','Accept':'gzip, deflate','token':gToken}
    if len(zipcode) == 5 :
        myparams   = {'country': 'USA', 'postalcode': zipcode}
    else :
        myparams   = {'country': 'CAN', 'postalcode': zipcode}
        
    #--------------------------------------------------------
    # send a GET request for the HEADENDS
    
    r0 = requests.get(HeadendsGetURL, headers=newHeaders, params=myparams)

    if r0.status_code != 200 :
        #logging.warning("SD:SDGetLineupPreview http(%d) SDcode(%d). Message: %s", r0.status_code, rj.get('code'), rj.get('message') )    
        logging.warning("SD:SDGetLineupPreview http(%d) ", r0.status_code )    
        return r0.status_code, None, None
    
    rj=r0.json()
    
    # search for a matching HeadEnd
    for lu in rj :
        if (lu['headend'] == zipcode) :
            logging.info('SD:SDGetLineupPreview: Found postal code %s', zipcode)
            break
        
    if (lu['headend'] != zipcode) :
        logging.warning("SD:SDGetLineupPreview : NO HEADEND for ZIP(%s)", zipcode)
#        return r0.status_code, None, None
        return 600, None, "NO HEADEND for ZIP(" + zipcode + ")"
    
    # if more than one lineup in this postalcode, then what????
    if ( len(lu['lineups']) > 1) :
        logging.warning("SD:SDGetLineupPreview: Multiple Antenna lineups for the ZIP.  Using LU 0.")
        
    #logging.warning("SDGetLineupPreview http(%d) SDcode(%d). Message: %s", r1.status_code, ys.get('code'), ys.get('message') )    
    
    # response could also be an error block
    # search the response list for headend=zip, transport= "Antenna"
    # lineups[0]['lineup'].
    #   lineup 'USA-OTA-53593'
    #   name '53593'
    #   uri '/20141201/lineups/USA-OTA-53593'
    
    # Get the Lineup Name
    sdlineup = lu['lineups'][0]['lineup'] 

    # Append the 'lineup' name to /lineups/preview/ to get a full preview URL
    requrl = PreviewLineupsGetURL + sdlineup
    
    # Send the requesst to SD to Get the Channel Lineup
    r1 = requests.get(requrl, headers=newHeaders)
    
    # Could get a 200 code with this json response
    # "response":"LINEUP_QUEUED"
    # "code": 2107
    # "serverID": 20141201.web.1"
    # "message":"The lineup is being generated, please retry."
    # "datetime":"dateTtimeZ"
    
    # Handle success first
    if r1.status_code == 200 :
        rj = r1.json()
        return r1.status_code, rj, sdlineup
    
    if r1.status_code == 500 :
        logging.warning("SD:SDGetLineupPreview http(%d) Message: %s", r1.status_code, r1.text )    
        return r1.status_code, None, None

    # handle some other error like 403???
    rj = r1.json()
    return r1.status_code, rj.get('code'), rj.get('message')


#==============================================================================

def SDGetProgramDetails(pidlist) :
    
    global gToken
    
    # To handle InProgress requests that occur near GMT=0, we need the previous days listings
    # This is a boundary condition but for the USA it affects 'Prime time.'
    # at CST of -5Z GMT 'today' starts ~7pm cst, so a 6pm(18:00) InProgress show is on
    # 'yesterdays' 23 GMT data
    
    if not gToken :  # Todo, this doesnt need to be here?
        SDGetToken()
   
    # header need the token
    newHeaders = {'User-Agent':PIGS_USER_AGENT, 'Content-type':'text/plain','Accept':'gzip, deflate','token':gToken}

    # ToDo what happens in no network connection?
    r1 = requests.post(ProgramDetailsPostURL, json=pidlist, headers=newHeaders)
  
    if r1.status_code == 200 :
        #logging.warning("SD:SDGetProgramDetails http(%d) SDcode(%d). Message: %s", r1.status_code, ys.get('code'), ys.get('message') )    
        return r1.status_code, r1.json()

    if r1.status_code == 500 :
        logging.warning("SD:SDGetProgramDetails http(%d) Message: %s", r1.status_code, r1.text )    
        return r1.status_code, None
    
    ys = r1.json
    logging.warning("SD:SDGetProgramDetails http(%d) SDcode(%d). Message: %s", r1.status_code, ys.get('code'), ys.get('message') )    
    rj = r1.json()
    return r1.status_code, rj.get('code'), rj.get('message')

#==============================================================================
# get data for all days available days=0, or n days if days >0 

from datetime import timedelta
from datetime import datetime
from time import gmtime, strftime, mktime

def SDGetStationSchedule(stationID, ndays) :

    global gToken
    
    # To handle InProgress requests that occur near GMT=0, we need the previous days listings
    # This is a boundary condition but for the USA it affects 'Prime time.'
    # at CST of -5Z GMT 'today' starts ~7pm cst, so a 6pm(18:00) InProgress show is on
    # 'yesterdays' 23 GMT data

    # recommended way to create an object representing the current time in UTC is by calling
    # t=datetime.utcnow()
    
    # build a request for yesterday and all days
   #t = datetime.now(timezone.utc)
    t  = datetime.utcnow()       # today
    td = t.strftime("%Y-%m-%d")  # td todays date as a string
    y  = t + timedelta(days=-1)  # yesterday is today -1 day
    yd = y.strftime("%Y-%m-%d")  # yd yesterday as a string
    
    # Sched Direct api allows for multiple station id's / dates in 1 call
    # I'm using this to request a single day / yesterday in 1 line
    # ommiting the date in line 2 requests all dates available
    # duplicating dates is handled by the SD server
    sl = [ {"stationID": stationID, 'date':[yd] },  # yesterday
           {"stationID": stationID } ]              # and everything else
    
    # for debugging, limit data to 'ndays' of program data
    # yes the dates are added in reverse order but SD does not seem to care
    if ndays>0 :
        # make a new sl object structure with 1 station, yesterday plus ndays specific dates
        sl = [ {"stationID": stationID, 'date':[yd] }]
        while ndays :
            dy = t + timedelta(days=ndays-1)
            sl[0]['date'].append( dy.strftime("%Y-%m-%d") )
            ndays -= 1          
           
    #sl = [ {"stationID": stationID, 'date':[td] }]  # uncomment for today only
    newHeaders = {'User-Agent':PIGS_USER_AGENT,'Content-type':'text/plain','Accept':'gzip, deflate','token':gToken}
 
    # ToDo what happens in no network connection?
    r1 = requests.post(SchedulesPostURL, json=sl, headers=newHeaders)

    # ToDo Could use the ScheudlesMd5 call to get all the available dates exept for yesterday
 
    logging.debug ("SD:SDGetStationSchedule Status code: %d", r1.status_code)

    ys = r1.json()
    # an outdated token would give a server error and no data
#    r.json()['code'] == 7020
#    r.json()['response'] == SCHEDULE_RANGE_EXCEEDED

    if r1.status_code == 200 :
        return r1.status_code, ys

    logging.warning("SD:SDGetStationSchedule http(%d) SDcode(%d). Message: %s", r1.status_code, ys.get('code'), ys.get('message') )    

    return r1.status_code, None

